#ifndef COSTCALCULATOR_H
#define COSTCALCULATOR_H

#include <functional>
#include <mutex>
#include <string>
#include <thread>
#include <utility>
#include <vector>

#include "commonconfig.h"
#include "icostcalculator.h"

namespace optimization
{
    /*!
     * \brief The CostFunctionsId enum holds the implemented one objective cost function.
     */
    enum class CostFunctionsId : unsigned
    {
        ackley,
        breakDownFrequency,
        constrainedMishrasBird,
        endToEndResponseTime,
        griewank,
        michalewicz,
        nqueens,
        rastrigin,
        robotMovement,
        rosenbrock,
        schwefel,
        utilization,
        utilizationConstEndToEnd
    };

    static const std::string ACKLEY("ACKLEY");                  /*!< Holds the ackley cost function option as a string. */
    static const std::string BREAKDOWNFREQUENCY("BREAKDOWNFREQUENCY"); /*!< Holds the frequency cost function as a string. */
    static const std::string CONSTRAINEDMISHRASBIRD("CONSTRAINEDMISHRASBIRD"); /*!< Holds the constrained mishra's bird cost function option as a string. */
    static const std::string EERT("EERT");                      /*!< Holds the EERT cost function option as a string. */
    static const std::string GRIEWANK("GRIEWANK");              /*!< Holds the griewank cost function option as a string. */
    static const std::string MICHALEWICZ("MICHALEWICZ");        /*!< Holds the utilization cost function option as a string. */
    static const std::string NQUEENS("NQUEENS");                /*!< Holds the n-queens problems cost function option as a string. */
    static const std::string RASTRIGIN("RASTRIGIN");            /*!< Holds the rastrigin cost function option as a string. */
    static const std::string ROBOTMOVEMENT("ROBOTMOVEMENT");    /*!< Holds the robotmovement cost function option as a string. */
    static const std::string ROSENBROCK("ROSENBROCK");          /*!< Holds the rosenbrock cost function option as a string. */
    static const std::string SCHWEFEL("SCHWEFEL");              /*!< Holds the schwefel cost function option as a string. */
    static const std::string UCEERT("UCEERT");                  /*!< Holds the utilization constrained EERT cost function option as a string. */
    static const std::string UTILIZATION("UTILIZATION");        /*!< Holds the utilization cost function option as a string. */

    /*!
     * \brief Gets the cost function list as a vector of strings.
     * \return A vector of strings with the implemented cost functions.
     */
//    std::vector<std::pair<std::string, std::string>>& getCostFunctionList()
//    {
//        static std::vector<std::pair<std::string, std::string>> costFunctionNames;
//        static std::once_flag flag;
//        std::call_once(flag, []()
//        {
//            costFunctionNames.clear();
//            costFunctionNames.push_back(std::make_pair(ACKLEY, "Ackley Function - Many Local Minima"));
//            costFunctionNames.push_back(std::make_pair(BREAKDOWNFREQUENCY, "Breakdown Frequency Analysis"));
//            costFunctionNames.push_back(std::make_pair(CONSTRAINEDMISHRASBIRD, "Constrained Mishra's Bird Function - Constraints"));
//            costFunctionNames.push_back(std::make_pair(EERT, "End-to-End Response Time Analysis"));
//            costFunctionNames.push_back(std::make_pair(GRIEWANK, "Griewank Function - Many Local Minima"));
//            costFunctionNames.push_back(std::make_pair(MICHALEWICZ, "Michalewicz Function - Steep Ridges/Drops"));
//            costFunctionNames.push_back(std::make_pair(NQUEENS, "N-Queens Problem Function - Discrete Problem"));
//            costFunctionNames.push_back(std::make_pair(RASTRIGIN, "Rastrigin Function - Many Local Minima"));
//            costFunctionNames.push_back(std::make_pair(ROBOTMOVEMENT, "Robot Movement Wrapper Cost Function"));
//            costFunctionNames.push_back(std::make_pair(ROSENBROCK, "Rosenbrock Function - Valley-Shaped"));
//            costFunctionNames.push_back(std::make_pair(SCHWEFEL, "Schwefel Function - Many Local Minima"));
//            costFunctionNames.push_back(std::make_pair(UCEERT, "End-to-End Response Time Analysis using Utilization Analysis as Constraint"));
//            costFunctionNames.push_back(std::make_pair(UTILIZATION, "Processors and Links Utilization Analysis"));
//        });
//        return costFunctionNames;
//    }

//    /*!
//     * \brief Converts cost function identification to its name as a string.
//     * \param id The cost function identification.
//     * \return The cost function name.
//     */
//    std::string convertCostFunctionsIdToString(CostFunctionsId id)
//    {
//        switch (id) {
//        case CostFunctionsId::ackley:
//            return ACKLEY;
//            break;
//        case CostFunctionsId::breakDownFrequency:
//            return BREAKDOWNFREQUENCY;
//            break;
//        case CostFunctionsId::constrainedMishrasBird:
//            return CONSTRAINEDMISHRASBIRD;
//            break;
//        case CostFunctionsId::endToEndResponseTime:
//            return EERT;
//            break;
//        case CostFunctionsId::griewank:
//            return GRIEWANK;
//            break;
//        case CostFunctionsId::michalewicz:
//            return MICHALEWICZ;
//            break;
//        case CostFunctionsId::nqueens:
//            return NQUEENS;
//            break;
//        case CostFunctionsId::rastrigin:
//            return RASTRIGIN;
//            break;
//        case CostFunctionsId::robotMovement:
//            return ROBOTMOVEMENT;
//            break;
//        case CostFunctionsId::rosenbrock:
//            return ROSENBROCK;
//            break;
//        case CostFunctionsId::schwefel:
//            return SCHWEFEL;
//            break;
//        case CostFunctionsId::utilization:
//            return UTILIZATION;
//            break;
//        case CostFunctionsId::utilizationConstEndToEnd:
//            return UCEERT;
//            break;
//        default:
//            break;
//        }
//        throw std::runtime_error(
//                    std::string("Error: optimization::convertCostFunctionsIdToString:") +
//                    std::to_string(__LINE__) +
//                    std::string(" given CostFunctionsId is not known."));
//    }

    /*!
     * \brief Converts cost function name to its identification.
     * \param name The cost function name.
     * \return The cost function identification.
     */
    CostFunctionsId convertCostFunctionsNameToId(const std::string& name)
    {
        CostFunctionsId type;
        if (name == ACKLEY) {
            type = CostFunctionsId::ackley;
        } else if (name == BREAKDOWNFREQUENCY) {
            type = CostFunctionsId::breakDownFrequency;
        } else if (name == CONSTRAINEDMISHRASBIRD) {
            type = CostFunctionsId::constrainedMishrasBird;
        } else if (name == EERT) {
            type = CostFunctionsId::endToEndResponseTime;
        } else if (name == GRIEWANK) {
            type = CostFunctionsId::griewank;
        } else if (name == MICHALEWICZ) {
            type = CostFunctionsId::michalewicz;
        } else if (name == NQUEENS) {
            type = CostFunctionsId::nqueens;
        } else if (name == RASTRIGIN) {
            type = CostFunctionsId::rastrigin;
        } else if (name == ROBOTMOVEMENT) {
            type = CostFunctionsId::robotMovement;
        } else if (name == ROSENBROCK) {
            type = CostFunctionsId::rosenbrock;
        } else if (name == SCHWEFEL) {
            type = CostFunctionsId::schwefel;
        } else if (name == UCEERT) {
            type = CostFunctionsId::utilizationConstEndToEnd;
        } else if (name == UTILIZATION) {
            type = CostFunctionsId::utilization;
        } else {
            throw std::runtime_error(
                        std::string("Error: optimization::convertCostFunctionsNameToId:") +
                        std::to_string(__LINE__) +
                        std::string(" given cost function name as a std::string is not known."));
        }
        return type;
    }

}

/*!
 * \brief The \c CostCalculator class represents an interface for any cost function calculator.
 * Where \c DecisionVariableType is the the data type for the cost calculator decision variable.
 * Where \c ConvertedType is the is the data type in which the decision variables have to be converted to in order to be processed.
 * Where \c Objective Type is the data type in which the cost calculator output is represented.
 */
template <typename DecisionVariableType, typename ConvertedType, typename ObjectiveType>
class CostCalculator : virtual public ICostCalculator
{
public:
    /*!
     * \brief Virtual Destructor.
     */
    virtual ~CostCalculator()
    {
    }

    /*!
     * \brief Sets the function responsible to convert the decision variable to the format to
     * a vector of integers, i.e. the placement representation used by the
     * \param fn A reference to the function responsible to convert from the decision variable format
     * to the system format.
     * \example In the case of representation of tasks placement used by the
     * system-on-chip \sa System, the type \c ConvertedType can be used as an integer type \c int.
     * \example In the case in which the problem search space is not homogeneus for the dimensions
     * the conversor can be used to expand or contract the decision variables to fit the
     * non-homogeneous search space.
     */
    virtual void setConversor(std::function<std::vector<ConvertedType>(std::vector<DecisionVariableType> &decisionVariable)> fn) = 0;

    /*!
     * \brief Performs a conversion from \c DecisionVariableType to \c ConvertedType for the
     * decision variable using the setted conversor function.
     * \param decisionVariable The cost function decision variable.
     * \return The converted decision variable.
     */
    virtual std::vector<ConvertedType> convertDecisionVariable(std::vector<DecisionVariableType> &decisionVariable) = 0;

    /*!
     * \brief Performs the evaluation.
     * \param decisionVariable The decision variable that encodes the task mapping placement.
     * \return A vector of results to the cost function.
     */
    virtual std::vector<ObjectiveType> calculateCostFunction(std::vector<DecisionVariableType> &decisionVariable) = 0;

    /*!
     * \brief Getter to the number of objectives returned by this cost function.
     * \return The number of objectives calculated by this cost function
     */
    virtual int getNumberOfObjectives() = 0;

    /*!
     * \brief Creates a clone of this object.
     * \return The clone reference.
     */
    virtual CostCalculator* clone() const = 0;
};

#endif // COSTCALCULATOR_H
