#ifndef NQUEENSCALCULATOR_H
#define NQUEENSCALCULATOR_H

#include <algorithm>
#include <functional>
#include <vector>

#include "commonconfig.h"
#include "costcalculator.h"

/*!
 * \brief The \c NQueensCalculator class calculates the N-queens problem cost function.
 * Given the positions of N queens chess pieces in a NxN board calculating the number of queens
 * threatened by each other.
 * \sa https://en.wikipedia.org/wiki/Eight_queens_puzzle
 * \sa U. Sarkar and S. Nag, “An Adaptive Genetic Algorithm for Solving N-Queens Problem,” 2017.
 */
template <typename DecisionVariableType>
class NQueensCalculator : public virtual CostCalculator<DecisionVariableType, int, _float>
{
    /*!
     * \brief Holds a reference to the function that converts the decision variables to
     * the task placement format.
     */
    std::function<std::vector<int>(
            std::vector<DecisionVariableType> &decisionVariable)> _conversor;

    /*!
     * \brief Holds the number of queens in the board. It also matches the size of the board.
     */
    unsigned _numberOfQueens;

public:
    /*!
     * \brief Default Constructor.
     * \param numberOfQueens Number of the Queens in the problem. It also represents the board's
     * side size.
     */
    NQueensCalculator(unsigned numberOfQueens) :
        _numberOfQueens(numberOfQueens)
    {
    }

    /*!
     * \brief Default Destructor
     */
    ~NQueensCalculator()
    {
        _conversor = nullptr;
    }

    /*!
     * \brief Sets the number of queens in the cost calculator.
     * \param numberOfQueens Number of queens in the problem.
     */
    void setNumberOfQueens(int numberOfQueens)
    {
        _numberOfQueens = numberOfQueens;
    }

    void setConversor(std::function<std::vector<int>(
                          std::vector<DecisionVariableType> &decisionVariable)> fn) override
    {
        _conversor = fn;
    }

    std::vector<int> convertDecisionVariable(
            std::vector<DecisionVariableType> &decisionVariable) override
    {
        return _conversor(decisionVariable);
    }

    std::vector<_float> calculateCostFunction(
            std::vector<DecisionVariableType> &decisionVariable) override
    {
        if (_numberOfQueens != decisionVariable.size()) {
            throw std::runtime_error(
                        std::string("Error: NQueensCalculator::calculateCostFunction:") +
                        std::to_string(__LINE__) +
                        std::string(" number of decision variables different than the"
                                    " number of queens."));
        }

        int leftDiagonals = 0;
        int rightDiagonals = 0;
        std::vector<int> positions = _conversor(decisionVariable);

        std::vector<int> f1(_numberOfQueens);
        std::vector<int> f2(_numberOfQueens);
        for (size_t i = 0; i < _numberOfQueens; i++) {
            f1[i] = positions[i] - i;
            f2[i] = (_numberOfQueens) - positions[i] - i;
        }

        std::sort(f1.begin(), f1.end());
        std::sort(f2.begin(), f2.end());
        for (size_t i = 1; i < _numberOfQueens; i++) {
            if (f1[i] == f1[i - 1]) {
                leftDiagonals++;
            }
            if (f2[i] == f2[i - 1]) {
                rightDiagonals++;
            }
        }
        
        std::vector<_float> objVals(1);
        objVals[0] = leftDiagonals + rightDiagonals;
        return objVals;
    }

    int getNumberOfObjectives() override
    {
        return 1;
    }

    CostCalculator<DecisionVariableType, int, _float>* clone() const override
    {
        return new NQueensCalculator<DecisionVariableType>(*this);
    }

    std::string getName() override
    {
        return std::string("NQueens");
    }
};

#endif // NQUEENSCALCULATOR_H
