#ifndef ENDTOENDSCHEDULABILITYCALCULATOR_H
#define ENDTOENDSCHEDULABILITYCALCULATOR_H

#include <functional>
#include <memory>
#include <string>

#include "commonconfig.h"
#include "costcalculator.h"
#include "system/system.h"

/*!
 * \brief The \c EndToEndSchedulabilityCalculator class represents a mono-objective cost function that
 * analyses the amount of unschedulable tasks and their flows using the end-to-end schedulability
 * analysis.
 * Where \c DecisionVariableType is the the data type representing the tasks placement in the set of processors.
 * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
 */
template <typename DecisionVariableType>
class EndToEndSchedulabilityCalculator : public virtual CostCalculator<DecisionVariableType, int, _float>
{
protected:
    /*!
     * \brief Holds a reference to the system-on-chip representation \c System.
     */
    std::shared_ptr<System> _system;

    /*!
     * \brief Holds a reference to the function that converts the decision variables to
     * the task placement format.
     */
    std::function<std::vector<int>(std::vector<DecisionVariableType> &decisionVariable)> _conversor;

    /*!
     * \brief Holds whether this cost function calculator is ready to be used.
     */
    bool _isReady;

public:
    /*!
     * \brief Constructor from a CSV file.
     * \note Note that using this constructor the platform used will have 4x4 cores.
     * \param applicationFile The file name of a text file inside of the default testbench directory
     * holding information about the real-time application.
     */
    EndToEndSchedulabilityCalculator(const std::string &applicationFile) :
        _system(std::make_shared<System>(std::make_shared<Application>(applicationFile))),
        _isReady(false)
    {
    }

    /*!
     * \brief Constructor from a system reference.
     * \param systemRef A reference to a system.
     */
    EndToEndSchedulabilityCalculator(std::shared_ptr<System> systemRef) :
        _system(systemRef),
        _isReady(false)
    {
    }

    /*!
     * \brief Copy Constructor.
     * \param other Other instance of this calculator.
     */
    EndToEndSchedulabilityCalculator(const EndToEndSchedulabilityCalculator &other) :
        _system(std::make_shared<System>(*other._system)),
        _conversor(other._conversor),
        _isReady(other._conversor)
    {
    }

    /*!
     * Default Destructor.
     */
    ~EndToEndSchedulabilityCalculator()
    {
        _system = nullptr;
    }

    void setConversor(std::function<std::vector<int>(std::vector<DecisionVariableType> &decisionVariable)> fn) override
    {
        _conversor = fn;
        _isReady = true;
    }

    std::vector<int> convertDecisionVariable(std::vector<DecisionVariableType> &decisionVariable) override
    {
        return _conversor(decisionVariable);
    }

    std::vector<_float> calculateCostFunction(std::vector<DecisionVariableType> &decisionVariable) override
    {
        std::vector<_float> objVals;
        if (_isReady) {
            std::vector<bool> analysisResults;
            _system->doMapping(_conversor(decisionVariable));
            _system->doEndToEndSchedulabityAnalysis(analysisResults);
            _float val = 0.0;
            for (bool isTaskUnschedulable : analysisResults) {
                val += isTaskUnschedulable ? 1.0 : 0.0;
            }
            objVals.push_back(val);
        }
        return objVals;
    }

    unsigned getNumberOfObjectives() const override
    {
        return 1;
    }

    CostCalculator<DecisionVariableType, int, _float>* clone() const override
    {
        return new EndToEndSchedulabilityCalculator<DecisionVariableType>(*this);
    }

    std::string getName() const override
    {
        return std::string("EERT");
    }

    std::pair<int, int> getDefaultSearchSpace() override
    {
        return std::make_pair(0, _system->getNumberOfPlatformProcessors() - 1);
    }

    unsigned getDimensions() const override
    {
        return _system->getNumberOfApplicationTasks();
    }
};

#endif // ENDTOENDSCHEDULABILITYCALCULATOR_H
