#ifndef APPLICATIONGENERATOR_H
#define APPLICATIONGENERATOR_H

#include "../commonconfig.h"
#include "application.h"
#include "flow.h"
#include "task.h"

#include <string>

/*!
 * \brief The \c ApplicationGenerator class is responsible for loading or saving tasks and flows from a
 * application from/in a CSV text.
 * \sa Application.
 */
class ApplicationGenerator
{
    /*!
     * \brief The FileAttributes enum holds the list of possible attributes that a file can have.
     */
    enum class FileAttribute {
        DEFAULT = 0,
        NAME = 1,
        CHAIN,
        COST,
        DEADLINE,
        PERIOD,
        DEST_NAME,
        PAYLOAD,
        PRIORITY,
        BLOCK,
        TASK_JITTER
    };

    /*!
     * \brief Holds the default directory path that holds testbenches.
     */
    static const std::string _defaultTestbenchDirPath;

    /*!
     * \brief Converts a string to a file attribute enumerate member.
     * \param attribute The attribute as a string.
     * \return The attribute as a \c FileAttribute.
     */
    FileAttribute _convertStringToFileAttribute(const std::string &attribute);

    /*!
     * \brief Converts a \c FileAttribute to a string.
     * \param attribute The attribute as a \c FileAttribute.
     * \return The attribute as a string.
     */
    const std::string _convertFileAttributeToString(FileAttribute attribute);

public:
    /*!
     * \brief Default Constructor.
     */
    ApplicationGenerator();

    /*!
     * \brief Generates application's tasks and flows from a CSV text file.
     * \param dirPath A string holding the directory path. If this string is empty the default directory is used instead.
     * \param fileName A string holding the file name.
     * \param tasks The map of tasks created by this method.
     * \param flows The map of flows creted by this method.
     */
    void generateTasksAndFlowsFromCsvFile(const std::string &dirPath, const std::string &fileName,
                                          std::map<std::string, std::shared_ptr<Task>> &tasks,
                                          std::map<std::string, std::shared_ptr<Flow>> &flows);

    /*!
     * \brief Generates a CSV text file from an application tasks and flows.
     * \param dirPath A string holding the directory path. If this string is empty the default directory is used instead.
     * \param fileName A string holding the file name.
     * \param tasks A vector of tasks that will be saved by this method.
     * \param flows A vector of flows that will be saved by this method.
     */
    void saveTasksAndFlowsToCsvFile(const std::string &dirPath, const std::string &fileName,
                                    std::vector<std::shared_ptr<Task>> &tasks,
                                    std::vector<std::shared_ptr<Flow> > &flows);

};

#endif // APPLICATIONGENERATOR_H
