#ifndef PROCESSOR_H
#define PROCESSOR_H

#include <memory>
#include <vector>

#include "../commonconfig.h"
#include "task.h"
#include "switcher.h"

/*!
 * \brief The Processor class represents a processing core in the
 * system-on-chip platform.
 * As described in:
 * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
 */
class Processor
{
    /*!
     * \brief Holds a counter to the total of processors.
     */
    static unsigned _processorCounter;

    /*!
     * \brief Holds an id number for this processor.
     */
    unsigned _id;

    /*!
     * \brief Holds the list of tasks that this processor will execute.
     */
    std::vector<std::shared_ptr<Task>> _tasks;

    /*!
     * \brief Holds a reference to the switcher element connected to this processor.
     */
    std::shared_ptr<Switcher> _switcher;

public:
    /*!
     * \brief Default Constructor.
     */
    Processor();

    /*!
     * \brief Copy Constructor.
     * \param other The other instance.
     */
    Processor(const Processor& other);

    /*!
     * \brief Default Destructor.
     */
    ~Processor();

    /*!
     * \brief Checks whether a task was mapped to this processor.
     * \param task A reference to a task.
     * \return True if this processor has this task mapped,
     * False otherwise.
     */
    bool contains(const std::shared_ptr<Task> task);

    /*!
     * \brief Calculates the utilization rate for this processor.
     * The result is also stored in the processor utilization rate.
     * This algorithm behaves as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \return This processor utilization rate.
     */
    _float calculateUtilization();

    /*!
     * \brief Maps a task to this processor.
     * \param task The task being mapped.
     */
    void mapTask(std::shared_ptr<Task> task);

    /*!
     * \brief Unmaps a task to this processor.
     * \param task the task being unmapped.
     */
    void unmapTask(std::shared_ptr<Task> task);

    /*!
     * \brief Maps multiple tasks to this processor.
     * \param tasks The tasks being mapped.
     */
    void mapTasks(std::vector<std::shared_ptr<Task>> tasks);

    /*!
     * \brief Unmaps all tasks in this processor.
     */
    void unmapAllTasks();

    /*!
     * \brief Connects this processor to a switcher.
     * \param switcher A reference to a switcher.
     */
    void connectToSwitcher(std::shared_ptr<Switcher> getSwitcher);

    /*!
     * \brief Getter to the switcher connected to this processor.
     * \return A reference to this processor's switcher.
     */
    std::shared_ptr<Switcher> getSwitcher() const;

    /*!
     * \brief Gets a vector of references to all tasks mapped to this processor.
     * \return A vector of reference to this processor's tasks.
     */
    const std::vector<std::shared_ptr<Task>> &getAllMappedTasks() const;
};

#endif // PROCESSOR_H
