#ifndef SWITCHER_H
#define SWITCHER_H

#include <memory>
#include <utility>
#include <vector>

#include "../commonconfig.h"
#include "link.h"



/*!
 * \brief The Switcher class represents a switching element for a network-on-chip.
 * That is basically a combination of a Network Interface Element and a Router.
 */
class Switcher
{
    /*!
     * \brief Holds a counter to the total number of switchers.
     */
    static unsigned _switcherCounter;

    /*!
     * \brief Holds an identification number to this switcher.
     */
    unsigned _id;

    /*!
     * \brief Holds the position of this switcher in the platform.
     */
    unsigned _address;

    /*!
     * \brief Holds an array of references to the links that connects this switcher.
     */
    std::vector<std::shared_ptr<Link>> _links;

    /*!
     * \brief Connects a pair of links to this switcher's router.
     * \param inputLink A reference to the input link.
     * \param outputLink A reference to the output link.
     * \param location The location where this pair of links will be connected.
     */
    void _connectLinks(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink,
                       LinkLocation location);

public:
    /*!
     * \brief Default Constructor.
     */
    Switcher();

    /*!
     * \brief Copy Constructor.
     * \param other Other Switcher instance.
     */
    Switcher(const Switcher& other);

    /*!
     * \brief Default Destructor.
     */
    ~Switcher();

    /*!
     * \brief Connects a pair of links to the local position of this switcher's router.
     * \param inputLink A reference to the input link.
     * \param outputLink A reference to the output link.
     */
    void connectLinksLocal(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink);

    /*!
     * \brief Connects a pair of links to the north position of this switcher's router.
     * \param inputLink A reference to the input link.
     * \param outputLink A reference to the output link.
     */
    void connectLinksNorth(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink);

    /*!
     * \brief Connects a pair of links to the south position of this switcher's router.
     * \param inputLink A reference to the input link.
     * \param outputLink A reference to the output link.
     */
    void connectLinksSouth(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink);

    /*!
     * \brief Connects a pair of links to the east position of this switcher's router.
     * \param inputLink A reference to the input link.
     * \param outputLink A reference to the output link.
     */
    void connectLinksEast(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink);

    /*!
     * \brief Connects a pair of links to the west position of this switcher's router.
     * \param inputLink A reference to the input link.
     * \param outputLink A reference to the output link.
     */
    void connectLinksWest(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink);

    /*!
     * \brief Gets the references to links connected in the local position.
     * \return A pair of references to the input and output links connected to this
     * switcher's local position.
     */
    std::pair<std::shared_ptr<Link>, std::shared_ptr<Link>> getLinksLocal();

    /*!
     * \brief Gets the references to links connected in the north position.
     * \return A pair of references to the input and output links connected to this
     * switcher's local position.
     */
    std::pair<std::shared_ptr<Link>, std::shared_ptr<Link>> getLinksNorth();

    /*!
     * \brief Gets the references to links connected in the south position.
     * \return A pair of references to the input and output links connected to this
     * switcher's local position.
     */
    std::pair<std::shared_ptr<Link>, std::shared_ptr<Link>> getLinksSouth();

    /*!
     * \brief Gets the references to links connected in the east position.
     * \return A pair of references to the input and output links connected to this
     * switcher's local position.
     */
    std::pair<std::shared_ptr<Link>, std::shared_ptr<Link>> getLinksEast();

    /*!
     * \brief Gets the references to links connected in the west position.
     * \return A pair of references to the input and output links connected to this
     * switcher's local position.
     */
    std::pair<std::shared_ptr<Link>, std::shared_ptr<Link>> getLinksWest();

    /*!
     * \brief Disconnects all links connected to this switcher.
     */
    void disconnectAllLinks();

    /*!
     * \brief Checks whether a link is connected to this switcher.
     * \param link A reference to a link that may be connected to this switcher.
     * \return true if the link is connected, false otherwise.
     */
    bool contains(std::shared_ptr<Link> link);

    /*!
     * \brief Getter to this switcher's address.
     */
    unsigned getAddress();

    /*!
     * \brief setAddress
     * \param address
     */
    void setAddress(const unsigned &address);
};

#endif // SWITCHER_H
