#ifndef UTILIZATIONSCHEDULABILITYCALCULATOR_H
#define UTILIZATIONSCHEDULABILITYCALCULATOR_H

#include <memory>
#include <string>
#include <functional>

#include "commonconfig.h"
#include "costcalculator.h"
#include "system/system.h"


/*!
 * \brief The UtilizationSchedulabilityCalculator class represents a mono-objective cost function
 * that counts how many processors under a specific mapping are running above their maximum capacity.
 * Where \c DecisionVariableType is the the data type representing the tasks placement in the set of processors.
 * \note The processor utilization test is a necessary test but not sufficient, i.e for a mapping be
 * successful it has to pass the utilization test, but mapping that passes this test are not necessarilly
 * schedulable.
 */
template <typename DecisionVariableType>
class UtilizationSchedulabilityCalculator : public virtual CostCalculator<DecisionVariableType, int, _float>
{
    /*!
     * \brief Holds a reference to the system-on-chip representation \c System.
     */
    std::shared_ptr<System> _system;

    /*!
     * \brief Holds a reference to the function that converts the decision variables to
     * the task placement format.
     */
    std::function<std::vector<int>(std::vector<DecisionVariableType> &decisionVariable)> _conversor;

    /*!
     * \brief Holds whether this cost function calculator is ready to be used.
     */
    bool _isReady;

public:
    /*!
     * \brief Constructor from a CSV file.
     * \param applicationFile The file name of a text file inside of the default testbench directory
     * holding information about the real-time application.
     */
    UtilizationSchedulabilityCalculator(const std::string &applicationFile) :
        _system(std::make_shared<System>(std::make_shared<Application>(applicationFile))),
        _isReady(false)
    {
    }

    /*!
     * \brief Constructor from a system reference.
     * \param systemRef A reference to the representation of the system.
     */
    UtilizationSchedulabilityCalculator(std::shared_ptr<System> systemRef) :
        _system(systemRef),
        _isReady(false)
    {
    }

    /*!
     * Default Destructor.
     */
    ~UtilizationSchedulabilityCalculator()
    {
        _system = nullptr;
    }

    void setConversor(std::function<std::vector<int>(std::vector<DecisionVariableType> &decisionVariable)> fn) override
    {
        _conversor = fn;
        _isReady = true;
    }

    std::vector<int> convertDecisionVariable(std::vector<DecisionVariableType> &decisionVariable) override
    {
        return _conversor(decisionVariable);
    }

    std::vector<_float> calculateCostFunction(std::vector<DecisionVariableType> &decisionVariable) override
    {
        std::vector<_float> objVals;
        if (_isReady) {
            _system->doMapping(_conversor(decisionVariable));
            std::vector<bool> procSchedulability;
            _system->doProcessorsUtilizationAnalysis(procSchedulability);
            _float val = 0.0;
            for (bool aboveCapacity : procSchedulability) {
                val += aboveCapacity ? 1.0 : 0.0;
            }
            std::vector<bool> linkSchedulability;
            _system->doLinksUtilizationAnalysis(linkSchedulability);
            for (bool aboveCapacity : linkSchedulability) {
                val += aboveCapacity ? 1.0 : 0.0;
            }
            objVals.push_back(val);
        }
        return objVals;
    }

    unsigned getNumberOfObjectives() const override
    {
        return 1;
    }

    CostCalculator<DecisionVariableType, int, _float>* clone() const override
    {
        return new UtilizationSchedulabilityCalculator<DecisionVariableType>(*this);
    }

    std::string getName() const override
    {
        return std::string("UTILIZATION");
    }

    std::pair<int, int> getDefaultSearchSpace()
    {
        return std::make_pair(0, _system->getNumberOfPlatformProcessors() - 1);
    }

    unsigned getDimensions() const override
    {
        return _system->getNumberOfApplicationTasks();
    }
};

#endif // UTILIZATIONSCHEDULABILITYCALCULATOR_H
