﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef COSTCALCULATOR_H
#define COSTCALCULATOR_H

#include <functional>
#include <mutex>
#include <string>
#include <thread>
#include <utility>
#include <vector>

#include "commonconfig.h"
#include "icostcalculator.h"

/*!
 * \brief The \c CostCalculator class represents an interface for any cost function calculator.
 * Where \c DecisionVariableType is the the data type for the cost calculator decision variable.
 * Where \c ConvertedType is the is the data type in which the decision variables have to be converted to in order to be processed.
 * Where \c Objective Type is the data type in which the cost calculator output is represented.
 */
template <typename DecisionVariableType, typename ConvertedType, typename ObjectiveType>
class CostCalculator : virtual public ICostCalculator
{
public:
    /*!
     * \brief Virtual Destructor.
     */
    virtual ~CostCalculator()
    {
    }

    /*!
     * \brief Sets the function responsible to convert the decision variable to the format to
     * a vector of integers, i.e. the placement representation used by the
     * \param fn A reference to the function responsible to convert from the decision variable format
     * to the system format.
     * \example In the case of representation of tasks placement used by the
     * system-on-chip \sa System, the type \c ConvertedType can be used as an integer type \c int.
     * \example In the case in which the problem search space is not homogeneus for the dimensions
     * the conversor can be used to expand or contract the decision variables to fit the
     * non-homogeneous search space.
     */
    virtual void setConversor(std::function<std::vector<ConvertedType>(std::vector<DecisionVariableType> &decisionVariable)> fn) = 0;

    /*!
     * \brief Performs a conversion from \c DecisionVariableType to \c ConvertedType for the
     * decision variable using the setted conversor function.
     * \param decisionVariable The cost function decision variable.
     * \return The converted decision variable.
     */
    virtual std::vector<ConvertedType> convertDecisionVariable(std::vector<DecisionVariableType> &decisionVariable) = 0;

    /*!
     * \brief Performs the evaluation.
     * \param decisionVariable The decision variable that encodes the task mapping placement.
     * \return A vector of results to the cost function.
     */
    virtual std::vector<ObjectiveType> calculateCostFunction(std::vector<DecisionVariableType> &decisionVariable) = 0;

    /*!
     * \brief Creates a clone of this object.
     * \return The clone reference.
     */
    virtual CostCalculator* clone() const = 0;

    /*!
     * \brief Getter to the homogeneous search space that this cost function is usually
     * evaluated.
     * \return A pair containing the minimum and maximum values for the dimensions in
     * a homogeneous search space.
     */
    virtual std::pair<ConvertedType, ConvertedType> getDefaultSearchSpace() = 0;
};

#endif // COSTCALCULATOR_H
