/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef APPLICATION_H
#define APPLICATION_H

#include <map>
#include <memory>
#include <string>
#include <vector>

#include "../commonconfig.h"
#include "flow.h"
#include "task.h"

/*!
 * \brief The \c Application class represents a real-time application with tasks that
 * can be executed in cores distributed in a system-on-chip platform.
 */
class Application
{
    /*!
     * \brief Holds a file name where this application model was loaded from.
     */
    std::string _fileName;

    /*!
     * \brief Holds a map of references for this application tasks with their names as keys.
     */
    std::map<std::string, std::shared_ptr<Task>> _tasks;

    /*!
     * \brief Holds an array of references for this application flows with their source task's name
     * as keys.
     */
    std::map<std::string, std::shared_ptr<Flow>> _flows;

    /*!
     * \brief Populates application's tasks and flows using a text file.
     * \param dirPath The directory path. If no directory is used it will use the default directory.
     * \param fileName The text file name.
     */
    void _populateTasksAndFlowsFromCsvFile(const std::string &fileName, const std::string &dirPath = std::string());

public:
    /*!
     * \brief Constructor of the application from a CSV text file.
     * \param fileName The file name of a text file inside of the default testbench directory.
     */
    Application(const std::string &fileName);

    /*!
     * \brief Constructor of the application from a CSV text file.
     * \param dirPath The directory path of the application text file.
     * \param fileName The file name of a text file inside of this directory.
     */
    Application(const std::string &dirPath, const std::string &fileName);

    /*!
     * \brief Copy Constructor.
     * \param other Other application instance.
     */
    Application(const Application &other);

    /*!
      * Default Destructor
      */
    ~Application();

    /*!
     * \brief Getter to the number of tasks in the application.
     * \return the number of tasks in the application.
     */
    unsigned getNumberOfTasks();

    /*!
     * \brief Gets a task from the application based on its name.
     * \param taskName The task name.
     * \return If this task exists returns a reference of a task.
     * Otherwise, returns \c nullptr.
     */
    std::shared_ptr<Task> getTask(const std::string &taskName);

    /*!
     * \brief Gets a map containing the reference of tasks in the application.
     * \return A map containing this application's tasks.
     */
    const std::map<std::string, std::shared_ptr<Task> > &getTasks() const;

    /*!
     * \brief Gets the application's tasks as \c std::vector.
     * \return A vector containing the application tasks' references.
     */
    std::vector<std::shared_ptr<Task>> getTasksAsVector();

    /*!
     * \brief Gets the application's flows as \c std::vector.
     * \return A vector containing the application flows' reference.
     */
    std::vector<std::shared_ptr<Flow>> getFlowsAsVector();

    /*!
     * \brief Gets the number of flows in this application..
     * \return The number of flows in the application.
     */
    unsigned getNumberOfFlows();

    /*!
     * \brief Gets a flow from the application based on its source task name.
     * \param taskName The source task name.
     * \return If the source task creates a communication flow returns a reference to this flow,
     * otherwise returns \c nullptr.
     */
    const std::shared_ptr<Flow> getFlow(const std::string &taskName);

    /*!
     * \brief Gets a map containing the reference of flows in the application.
     * \return A map containing this application's flows.
     */
    const std::map<std::string, std::shared_ptr<Flow> > &getFlows() const;

    /*!
     * \brief Gets the application name as a string.
     * \return The application name.
     */
    std::string getName() const;
};

#endif // APPLICATION_H
