﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "applicationgenerator.h"

#include <fstream>
#include <iostream>
#include <map>
#include <memory>
#include <sstream>

#include "../commonconfig.h"

ApplicationGenerator::FileAttribute ApplicationGenerator::_convertStringToFileAttribute(
        const std::string &attribute)
{
    FileAttribute ret = FileAttribute::DEFAULT;
    if (attribute == "NAME") {
        ret = FileAttribute::NAME;
    } else if (attribute == "CHAIN") {
        ret = FileAttribute::CHAIN;
    } else if (attribute == "COST") {
        ret = FileAttribute::COST;
    } else if (attribute == "DEADLINE") {
        ret = FileAttribute::DEADLINE;
    } else if (attribute == "PERIOD") {
        ret = FileAttribute::PERIOD;
    } else if (attribute == "DEST_NAME") {
        ret = FileAttribute::DEST_NAME;
    } else if (attribute == "PAYLOAD") {
        ret = FileAttribute::PAYLOAD;
    } else if (attribute == "PRIORITY") {
        ret = FileAttribute::PRIORITY;
    } else if (attribute == "BLOCK") {
        ret = FileAttribute::BLOCK;
    } else if (attribute == "TASK_JITTER") {
        ret = FileAttribute::TASK_JITTER;
    }

    return ret;
}

const std::string ApplicationGenerator::_convertFileAttributeToString(
        FileAttribute attribute)
{
    switch (attribute) {
    case FileAttribute::NAME:
        return "NAME";
        break;
    case FileAttribute::CHAIN:
        return "CHAIN";
        break;
    case FileAttribute::COST:
        return "COST";
        break;
    case FileAttribute::DEADLINE:
        return "DEADLINE";
        break;
    case FileAttribute::PERIOD:
        return "PERIOD";
        break;
    case FileAttribute::DEST_NAME:
        return "DEST_NAME";
        break;
    case FileAttribute::PAYLOAD:
        return "PAYLOAD";
        break;
    case FileAttribute::PRIORITY:
        return "PRIORITY";
        break;
    case FileAttribute::BLOCK:
        return "BLOCK";
        break;
    case FileAttribute::TASK_JITTER:
        return "TASK_JITTER";
    default:
        return "DEFAULT";
        break;
    }
}

const std::string ApplicationGenerator::_defaultTestbenchDirPath =
        std::string("data/testbench");

ApplicationGenerator::ApplicationGenerator()
{
}

void ApplicationGenerator::generateTasksAndFlowsFromCsvFile(const std::string &dirPath,
                                                            const std::string &fileName,
                                                            std::map<std::string, std::shared_ptr<Task>> &tasks,
                                                            std::map<std::string, std::shared_ptr<Flow>> &flows)
{
    std::string filePath = (dirPath.empty() ? _defaultTestbenchDirPath : dirPath) + "/" + fileName;

    std::ifstream inputDataReader;
    inputDataReader.open(filePath);
    if (!inputDataReader.is_open()) {
        throw std::runtime_error(ERROR_MSG(__FILE__, __LINE__,
                                           "CSV file ( " + filePath + " ) not found."));
    }

    // Read the first line to check it's attributes order
    std::string line;
    std::getline(inputDataReader, line);
    if (line.empty()) {
        throw std::runtime_error(ERROR_MSG(__FILE__, __LINE__,
                                           "CSV file ( " + filePath + " ) is empty."));
    }

    // Read the attributes order
    std::stringstream firstLineStream(line);
    std::string attr;
    std::vector<FileAttribute> attributesMap;
    for (int i = 0; std::getline(firstLineStream, attr, ','); i++) {
        attributesMap.push_back(_convertStringToFileAttribute(attr));
    }

    // Read Other Lines
    while (std::getline(inputDataReader, line)) {
        std::shared_ptr<Task> task = nullptr;
        std::shared_ptr<Flow> flow = nullptr;
        std::stringstream lineStream(line);
        for (int i = 0; std::getline(lineStream, attr, ','); i++) {
            switch (attributesMap.at(i)) {
            case FileAttribute::NAME:
                if (tasks.find(attr) == tasks.end()) {
                    task = std::make_shared<Task>();
                } else {
                    task = tasks[attr];
                }
                task->setName(attr);
                tasks[attr] = task;
                break;
            case FileAttribute::CHAIN:
                if (task) {
                    task->setChainName(attr);
                }
                break;
            case FileAttribute::COST:
                if (task) {
                    double cost = std::stod(attr);
                    task->setCost((cost > UNIVERSE_AGE_IN_SEC) ?
                                      UNIVERSE_AGE_IN_SEC : static_cast<_float>(cost));
                }
                break;
            case FileAttribute::DEADLINE:
                if (task) {
                    double deadline = std::stod(attr);
                    task->setDeadline((deadline > UNIVERSE_AGE_IN_SEC) ?
                                          UNIVERSE_AGE_IN_SEC : static_cast<_float>(deadline));
                }
                break;
            case FileAttribute::PERIOD:
                if (task) {
                    double period = std::stod(attr);
                    task->setPeriod((period > UNIVERSE_AGE_IN_SEC) ?
                                        UNIVERSE_AGE_IN_SEC : static_cast<_float>(period));
                }
                break;
            case FileAttribute::DEST_NAME:
                if (task && !attr.empty()) {
                    std::shared_ptr<Task> taskDst = nullptr;
                    if (tasks.find(attr) != tasks.end()) {
                        taskDst = tasks[attr];
                    } else {
                        taskDst = std::make_shared<Task>();
                        taskDst->setName(attr);
                        tasks[attr] = taskDst;

                    }
                    flow = std::make_shared<Flow>(task, taskDst);
                    flows[task->getName()] = flow;
                    task->setFlow(flow);
                }
                break;
            case FileAttribute::PAYLOAD:
                if (flow) {
                    flow->setFlowSize(static_cast<unsigned>(std::stod(attr)));
                }
                break;
            case FileAttribute::PRIORITY:
                if (task) {
                    task->setPriority(static_cast<unsigned>(std::stod(attr)));
                }
                break;
            case FileAttribute::TASK_JITTER:
                if (task) {
                    task->setOffsetJitter(static_cast<_float>(std::stod(attr)));
                }
                break;
            case FileAttribute::BLOCK:
            default:
                break;
            }
        }
    }
}

void ApplicationGenerator::saveTasksAndFlowsToCsvFile(const std::string &dirPath,
                                                      const std::string &fileName,
                                                      std::vector<std::shared_ptr<Task>> &tasks,
                                                      std::vector<std::shared_ptr<Flow>> &flows)
{
    std::string filePath = (dirPath.empty() ? _defaultTestbenchDirPath : dirPath) + "/" + fileName;

    std::ofstream outputDataWriter;
    outputDataWriter.open(filePath);
    if (!outputDataWriter.is_open()) {
        throw std::runtime_error(ERROR_MSG(__FILE__, __LINE__,
                                           "file to write application."));
        return;
    }

    ignore(flows);

    // Writes the first line with the attributes order
    std::vector<FileAttribute> attributesOrder(
                                       {FileAttribute::NAME, FileAttribute::CHAIN,
                                        FileAttribute::COST, FileAttribute::DEADLINE,
                                        FileAttribute::PERIOD, FileAttribute::DEST_NAME,
                                        FileAttribute::PAYLOAD, FileAttribute::PRIORITY});
    unsigned i;
    for (i = 0; i < attributesOrder.size() - 1; i++) {
        outputDataWriter << _convertFileAttributeToString(attributesOrder[i]) << ",";
    }
    outputDataWriter << _convertFileAttributeToString(attributesOrder[i]) << "\n";

    // writes other lines
    for (auto task : tasks) {
        for (FileAttribute attribute : attributesOrder) {
            switch (attribute) {
            case FileAttribute::NAME:
                outputDataWriter << task->getName() << ",";
                continue;
            case FileAttribute::CHAIN:
                outputDataWriter << task->getChainName() << ",";
                continue;
            case FileAttribute::COST:
                outputDataWriter << task->getCost() << ",";
                continue;
            case FileAttribute::DEADLINE:
                outputDataWriter << task->getDeadline() << ",";
                continue;
            case FileAttribute::PERIOD:
                outputDataWriter << task->getPeriod() << ",";
                continue;
            case FileAttribute::DEST_NAME:
                if (task->getFlow()) {
                    std::shared_ptr<Task> dstTask =
                            task->getFlow()->getDestinationTask().lock();
                    if (dstTask) {
                        outputDataWriter << dstTask->getName();
                    }
                }
                outputDataWriter << ",";
                continue;
            case FileAttribute::PAYLOAD:
                if (task->getFlow()) {
                    outputDataWriter << task->getFlow()->getFlowSize();
                }
                outputDataWriter << ",";
                continue;
            case FileAttribute::PRIORITY:
                outputDataWriter << task->getPriority();
                continue;
            default:
                continue;
            }
        }
        outputDataWriter << "\n";
    }
    outputDataWriter.close();
}
