/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef APPLICATIONGENERATOR_H
#define APPLICATIONGENERATOR_H

#include "../commonconfig.h"
#include "application.h"
#include "flow.h"
#include "task.h"

#include <string>

/*!
 * \brief The \c ApplicationGenerator class is responsible for loading or saving tasks and flows from a
 * application from/in a CSV text.
 * \sa Application.
 */
class ApplicationGenerator
{
    /*!
     * \brief The FileAttributes enum holds the list of possible attributes that a file can have.
     */
    enum class FileAttribute {
        DEFAULT = 0,
        NAME = 1,
        CHAIN,
        COST,
        DEADLINE,
        PERIOD,
        DEST_NAME,
        PAYLOAD,
        PRIORITY,
        BLOCK,
        TASK_JITTER
    };

    /*!
     * \brief Holds the default directory path that holds testbenches.
     */
    static const std::string _defaultTestbenchDirPath;

    /*!
     * \brief Converts a string to a file attribute enumerate member.
     * \param attribute The attribute as a string.
     * \return The attribute as a \c FileAttribute.
     */
    FileAttribute _convertStringToFileAttribute(const std::string &attribute);

    /*!
     * \brief Converts a \c FileAttribute to a string.
     * \param attribute The attribute as a \c FileAttribute.
     * \return The attribute as a string.
     */
    const std::string _convertFileAttributeToString(FileAttribute attribute);

public:
    /*!
     * \brief Default Constructor.
     */
    ApplicationGenerator();

    /*!
     * \brief Generates application's tasks and flows from a CSV text file.
     * \param dirPath A string holding the directory path. If this string is empty the default directory is used instead.
     * \param fileName A string holding the file name.
     * \param tasks The map of tasks created by this method.
     * \param flows The map of flows creted by this method.
     */
    void generateTasksAndFlowsFromCsvFile(const std::string &dirPath, const std::string &fileName,
                                          std::map<std::string, std::shared_ptr<Task>> &tasks,
                                          std::map<std::string, std::shared_ptr<Flow>> &flows);

    /*!
     * \brief Generates a CSV text file from an application tasks and flows.
     * \param dirPath A string holding the directory path. If this string is empty the default directory is used instead.
     * \param fileName A string holding the file name.
     * \param tasks A vector of tasks that will be saved by this method.
     * \param flows A vector of flows that will be saved by this method.
     */
    void saveTasksAndFlowsToCsvFile(const std::string &dirPath, const std::string &fileName,
                                    std::vector<std::shared_ptr<Task>> &tasks,
                                    std::vector<std::shared_ptr<Flow> > &flows);

};

#endif // APPLICATIONGENERATOR_H
