﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "flow.h"

#include <algorithm>
#include <cmath>
#include <iostream>
#include <map>
#include <set>

#include "mapper.h"
#include "link.h"
#include "platform.h"
#include "task.h"

unsigned Flow::_flowCounter = 0;

unsigned Flow::getUniqueId() const
{
    return _id;
}

std::weak_ptr<Task> Flow::getSourceTask() const
{
    return _srcTask;
}

std::weak_ptr<Task> Flow::getDestinationTask() const
{
    return _dstTask;
}

int Flow::getFlowSize() const
{
    return _flowSize;
}

void Flow::setFlowSize(int flowSize)
{
    _flowSize = flowSize;
    _flitSize = (flowSize / Platform::getFlitSizeInBits()) + 1; // N Flits plus the Header
}

int Flow::getFlitSize() const
{
    return _flitSize;
}

_float Flow::getBasicLatency() const
{
    return _basicLatency;
}

_float Flow::getWorstCaseLatency()
{
    return _worstCaseLatency;
}

std::vector<std::shared_ptr<Flow> > Flow::getDirectInterferenceArray() const
{
    return _directInterferenceArray;
}

std::vector<std::shared_ptr<Flow> > Flow::getIndirectInterferenceArray() const
{
    return _indirectInterferenceArray;
}

_float Flow::calculateBasicLatency(const std::vector<std::shared_ptr<Link> > &links)
{
    return _calculateBasicLatency(links);
}

_float Flow::calculateWorstCaseLatency()
{
    return _calculateWorstCaseLatency();
}

bool Flow::calculateDirectInterferenceSet(const std::vector<std::shared_ptr<Link> > &links)
{
    _directInterferenceArray.clear();
    // Check whether the flow doesn't use any link.
    if (!links.empty()) {
        auto task = getSourceTask().lock();
        if (!task) {
            throw std::runtime_error(
                    ERROR_MSG(__FILE__, __LINE__,
                              "source task refence is invalid."));
        }

        std::set<std::shared_ptr<Flow>> directInterferenceSet;
        for (auto link : links) {
            for (auto flow : link->getAllFlows()) {
                auto task_j = flow->getSourceTask().lock();
                if (!task_j) {
                    throw std::runtime_error(
                            ERROR_MSG(__FILE__, __LINE__,
                                      "source task_j refence is invalid."));
                }
                if (task_j->getPriority() <= task->getPriority() && task != task_j) {
                    directInterferenceSet.emplace(flow);
                }
            }
        }

        _directInterferenceArray.assign(directInterferenceSet.begin(),
                                        directInterferenceSet.end());
    }
    return true;
}

bool Flow::calculateIndirectInterferenceSet(const std::vector<std::shared_ptr<Flow> > &allFlows,
                                            Mapper &mapper)
{
    auto task = getSourceTask().lock();
    if (!task) {
        throw std::runtime_error(
                ERROR_MSG(__FILE__, __LINE__,
                          "source task refence is invalid."));
    }

    _indirectInterferenceArray.clear();
    if (!_directInterferenceArray.empty()) {
        for (auto flow_i : allFlows) {
            auto task_i = flow_i->getSourceTask().lock();
            if (!task_i) {
                throw std::runtime_error(
                        ERROR_MSG(__FILE__, __LINE__,
                                  "source task_i refence is invalid."));
            }

            if (task_i->getPriority() <= task->getPriority() && task_i != task &&
                    std::find(_directInterferenceArray.begin(), _directInterferenceArray.end(), flow_i)
                    == _directInterferenceArray.end()) {
                auto links_i = mapper.mapFlow(flow_i);
                for (auto flow_j : _directInterferenceArray) {
                    auto links_j = mapper.mapFlow(flow_j);
                    bool hasIntersection = std::find_first_of(links_i.begin(), links_i.end(),
                                                              links_j.begin(), links_j.end()) != links_i.end();
                    if (hasIntersection) {
                        _indirectInterferenceArray.push_back(flow_i);
                    }
                }
            }
        }
    }

    return true;
}


_float Flow::_calculateBasicLatency(const std::vector<std::shared_ptr<Link> > &links)
{
    unsigned numberOfLinks = unsigned(links.size());

    _basicLatency = !numberOfLinks ? 0.0 :
                                     numberOfLinks * Platform::getLatencyLinkHop() +
                                     (numberOfLinks - 1) * Platform::getLatencyRouterHop() +
                                     (std::ceil((_float(_flowSize) / Platform::getFlitSizeInBits())) *
                                      Platform::getLatencyLinkHop());
    return _basicLatency;
}

_float Flow::_calculateWorstCaseLatency()
{
    if (_basicLatency  < 0) {
        std::cerr <<
                     ERROR_MSG(__FILE__, __LINE__,
                               "basic latency is not calculated previously.") << std::endl;
        return -1.0;
    }

    auto task = getSourceTask().lock();
    if (!task) {
        throw std::runtime_error(
                ERROR_MSG(__FILE__, __LINE__,
                          "source task reference is invalid."));
    }

    _float D_i = task->getDeadline();
    _float L_i = getBasicLatency();
    _float I_n;
    _float S_n = 0.0;
    _float S_n_1 = L_i;

    // Calculate Indirect Jitters
    std::map<std::shared_ptr<Flow>, _float> indirectJitters;
    for (unsigned i = 0; i < _directInterferenceArray.size(); i++) {
        std::vector<std::shared_ptr<Flow>> directInterferenceArray_j =
                _directInterferenceArray.at(i)->getDirectInterferenceArray();
        if (!directInterferenceArray_j.empty() &&
                std::find_first_of(directInterferenceArray_j.begin(),
                                   directInterferenceArray_j.end(),
                                   _indirectInterferenceArray.begin(),
                                   _indirectInterferenceArray.end()) !=
                directInterferenceArray_j.end()) {
            if (_directInterferenceArray[i]->getWorstCaseLatency() < 0) {
                _directInterferenceArray[i]->calculateWorstCaseLatency();
            }
            indirectJitters[_directInterferenceArray[i]] =
                    (_directInterferenceArray[i]->getWorstCaseLatency() -
                     _directInterferenceArray[i]->getBasicLatency());
        } else {
            indirectJitters[_directInterferenceArray[i]] = 0.0;
        }
    }

    while (true) {
        // Update
        S_n = S_n_1;

        // Calculates Interference
        I_n = 0.0;
        for (auto flow_j : _directInterferenceArray) {
            auto task_j = flow_j->getSourceTask().lock();
            if (!task_j) {
                throw std::runtime_error(
                        ERROR_MSG(__FILE__, __LINE__,
                                  "source task_j refence is invalid."));
            }
            I_n += std::ceil((S_n + task_j->getResponseTime() +
                              indirectJitters[flow_j]) / task_j->getPeriod()) *
                    flow_j->getBasicLatency();
        }

        // New WCL
        S_n_1 = L_i + I_n;

        // Check to exit loop
        if ((S_n_1 - S_n) > TIME_EPSILON && S_n_1 <= D_i) {
            continue;
        } else {
            break;
        }
    }

    _worstCaseLatency = (S_n_1 > D_i) ? UNIVERSE_AGE_IN_SEC : S_n;

    return _worstCaseLatency;
}

Flow::Flow(const std::shared_ptr<Task> &srcTask, const std::shared_ptr<Task> &dstTask) :
    _id(++_flowCounter),
    _srcTask(srcTask),
    _dstTask(dstTask),
    _flowSize(0),
    _flitSize(0),
    _basicLatency(-1.0),
    _worstCaseLatency(-1.0)
{
    if (!_srcTask.lock()) {
        _srcTask = std::shared_ptr<Task>(new Task());
    }
}

Flow::Flow(const Flow &other) :
    _id(++_flowCounter),
    _srcTask(other._srcTask),
    _dstTask(other._dstTask),
    _flowSize(other._flowSize),
    _flitSize(other._flitSize),
    _basicLatency(other._basicLatency),
    _worstCaseLatency(other._basicLatency)
{
    for (const auto& flowd : other._directInterferenceArray) {
        _directInterferenceArray.push_back(std::make_shared<Flow>(*flowd));
    }

    for (const auto& flowi : other._indirectInterferenceArray) {
        _indirectInterferenceArray.push_back(std::make_shared<Flow>(*flowi));
    }
}

Flow::~Flow()
{
    _flowCounter--;

    for (auto &flow : _indirectInterferenceArray) {
        flow = nullptr;
    }
    _indirectInterferenceArray.clear();

    for (auto &flow : _directInterferenceArray) {
        flow = nullptr;
    }
    _directInterferenceArray.clear();
}
