/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef FLOW_H
#define FLOW_H

#include <memory>
#include <string>
#include <vector>

#include "../commonconfig.h"

class Mapper;

class Link;

class Task;

/*!
 * \brief The Flow class represents a traffic flow created by
 * a task communicating using the noc architecture.
 */
class Flow
{
    /*!
     * \brief Holds a counter to total number of flows.
     */
    static unsigned _flowCounter;

    /*!
     * \brief Holds the unique flow's id.
     */
    const unsigned _id;

    /*!
     * \brief Holds a reference to this flow's source task.
     */
    std::weak_ptr<Task> _srcTask;

    /*!
     * \brief Holds a reference to this flow's destination task.
     */
    std::weak_ptr<Task> _dstTask;

    /*!
     * \brief Holds this traffic flow in bits.
     */
    int _flowSize;

    /*!
     * \brief Holds this traffic flow in flits number.
     */
    int _flitSize;

    /*!
     * \brief Holds this traffic flow basic latency.
     * As default when not calculated, it holds -1.0.
     */
    _float _basicLatency;

    /*!
     * \brief Holds this traffic worst case latency.
     * As default when not calculated, it holds -1.0.
     */
    _float _worstCaseLatency;

    /*!
     * \brief Holds a vector to references of flows that
     * directly interfere with this flow communication.
     */
    std::vector<std::shared_ptr<Flow>> _directInterferenceArray;

    /*!
     * \brief Holds a vector to references of flows that
     * indirectly interfere with this flow communication.
     */
    std::vector<std::shared_ptr<Flow>> _indirectInterferenceArray;

    /*!
     * \brief Calculates the basic latency for this flow transmission as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * and
     * \sa Shi, Z., & Burns, A. (2008). Real-Time Communication Analysis for On-Chip Networks with Wormhole Switching. In ACM/IEEE International Symposium on Networks-on-Chip (pp. 161–170). https://doi.org/10.1109/NOCS.2008.4492735
     * \note It also hold the result to be used directly without new calculations \sa getBasicLatency.
     * \param links An array of links that this flow crosses during its transmission.
     * \note This method considers that the flow's size is one flit bigger to consider the header flit
     * for the message packet.
     * \return The calculated value for the basic latency.
     */
    _float _calculateBasicLatency(const std::vector<std::shared_ptr<Link>> &links);

    /*!
     * \brief Calculates the worst case latency for a flow transmission as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \note It also hold the result to be used directly without new calculations \sa getWorstCaseLatency.
     * \note This method considers that both direct and indirect interference and basic latency was
     * already calculated and the source task has its response time calculated previously too.
     * \return The calculated value for the worst case latency.
     */
    _float _calculateWorstCaseLatency();

public:
    /*!
     * \brief Default Constructor.
     * \param srcTask Flow's source.
     * \param dstTask Flow's destination.
     */
    Flow(const std::shared_ptr<Task> &getSourceTask, const std::shared_ptr<Task> &getDestinationTask);

    /*!
     * \brief Copy Constructor.
     * \param other Other instance of Flow.
     */
    Flow(const Flow &other);

    /*!
     * \brief Default Destructor.
     */
    ~Flow();

    /*!
     * \brief Getter to flow's unique id.
     */
    unsigned getUniqueId() const;

    /*!
     * \brief Getter to flow's source task reference.
     * \return The flow's source task reference.
     */
    std::weak_ptr<Task> getSourceTask() const;

    /*!
     * \brief Getter to flow's destination task reference.
     * \return The flow's destination task reference.
     */
    std::weak_ptr<Task> getDestinationTask() const;

    /*!
     * \brief Getter to flow's size in bits.
     * \return The flow's size in bits.
     */
    int getFlowSize() const;

    /*!
     * \brief Setter to flow's size in bits.
     * \param The flow's size in bits.
     */
    void setFlowSize(int getFlowSize);

    /*!
     * \brief Getter to flow's size in flits.
     * \return The flow's size in flits. The size in flits for a message is expressed by
     * the size in bits divided by the size of a flit in bits plus one, due to the header flit.
     */
    int getFlitSize() const;

    /*!
     * \brief Getter to flow's basic latency.
     * \return If the basic latency has not been calculated yet this method returns -1.0,
     * otherwise it returns the basic flow latency.
     */
    _float getBasicLatency() const;

    /*!
     * \brief Getter to flow's worst case latency.
     * \return If the worst case latency has not been calculated yet this method returns -1.0,
     * otherwise it returns the worst case flow latency.
     */
    _float getWorstCaseLatency();

    /*!
     * \brief Getter to array of flows that directly interfere with this flow.
     * \return Array of flows that directly interfere with this flow.
     */
    std::vector<std::shared_ptr<Flow>> getDirectInterferenceArray() const;

    /*!
     * \brief Getter to array of flows that indirectly interfere with this flow.
     * \return Array of flows  that indirectly interfere with this flow.
     */
    std::vector<std::shared_ptr<Flow>> getIndirectInterferenceArray() const;

    /*!
     * \brief Calculates the basic latency for this flow transmission as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * and
     * \sa Shi, Z., & Burns, A. (2008). Real-Time Communication Analysis for On-Chip Networks with Wormhole Switching. In ACM/IEEE International Symposium on Networks-on-Chip (pp. 161–170). https://doi.org/10.1109/NOCS.2008.4492735
     * \note It also hold the result to be used directly without new calculations \sa getBasicLatency.
     * \param links An array of links that this flow crosses during its transmission.
     * \note This method considers that the flow's size is one flit bigger to consider the header flit
     * for the message packet.
     * \return The calculated value for the basic latency.
     */
    _float calculateBasicLatency(const std::vector<std::shared_ptr<Link>> &links);

    /*!
     * \brief Calculates the worst case latency for a flow transmission as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \note It also hold the result to be used directly without new calculations \sa getWorstCaseLatency.
     * \note This method considers that both direct and indirect interference and basic latency was
     * already calculated and the source task has its response time calculated previously too.
     * \return The calculated value for the worst case latency.
     */
    _float calculateWorstCaseLatency();

    /*!
     * \brief Calculates the set of flows that directly interfere with this flow as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \param links An array of links that this flow crosses during its transmission.
     * \note Once the direct interferene set is calculated it can be accessed without new
     * caulcations using the method \sa getDirectInterferenceArray.
     * \return true if calculates sucessful, false otherwise.
     */
    bool calculateDirectInterferenceSet(const std::vector<std::shared_ptr<Link>> &links);

    /*!
     * \brief Calculates the set of flows that indirectly interfere with this flow as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \param flows An array with all posible flows that can interfere indirectly this flow.
     * \note Once the indirect interence set is calculated it can be accessed without new
     * calculations using the method \sa getDirectInterferenceArray.
     * \note This method considers that the direct interference was calculated before
     * \sa calculateDirectInterferenceSet.
     * \return true if calculates successful, false otherwise.
     */

    bool calculateIndirectInterferenceSet(const std::vector<std::shared_ptr<Flow>> &allFlows,
                                          Mapper &mapper);
};

#endif // FLOW_H
