﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "link.h"

#include <iostream>

#include "task.h"

unsigned Link::_linkCounter = 0;

LinkLocation Link::getLocation()
{
    return _location;
}

void Link::connectToSourceSwitcher(std::shared_ptr<Switcher> sourceSwitcher, LinkLocation location)
{
    _sourceSwitcher = sourceSwitcher;
    _location = location;
}

void Link::connectToDestinationSwitcher(std::shared_ptr<Switcher> destinationSwitcher)
{
    _destinationSwitcher = destinationSwitcher;
}

std::weak_ptr<Switcher> Link::getSourceSwitcher() const
{
    return _sourceSwitcher;
}

std::weak_ptr<Switcher> Link::getDestinationSwitcher() const
{
    return _destinationSwitcher;
}

Link::Link() :
    _id(++_linkCounter)
{

}

Link::Link(const Link &other) :
    _id(++_linkCounter),
    _location(other._location),
    _sourceSwitcher(other._sourceSwitcher),
    _destinationSwitcher(other._destinationSwitcher)
{
    for (const auto &flow : other._flows) {
        _flows.push_back(std::make_shared<Flow>(*flow));
    }
}

Link::~Link()
{
    _linkCounter--;

    for (auto &flow : _flows) {
        flow = nullptr;
    }
    _flows.clear();
}

bool Link::contains(const std::shared_ptr<Flow> flow)
{
    bool ret = false;
    for (const std::shared_ptr<Flow> flowInside : _flows) {
        if (flowInside == flow) {
            ret = true;
            break;
        }
    }
    return ret;
}

_float Link::calculateUtilization()
{
    _float utilization = 0.0;
    for (auto flow : _flows) {
        std::shared_ptr<Task> task = flow->getSourceTask().lock();
        if (!task) {
            throw std::runtime_error(
                        std::string("Error: Link::calculateUtilization:") +
                        std::to_string(__LINE__) +
                        std::string(" source task refence is invalid."));
        }
        utilization += flow->getBasicLatency() / task->getPeriod();
    }
    return utilization;
}

void Link::addFlow(std::shared_ptr<Flow> flow)
{
    _flows.push_back(flow);
}

void Link::removeFlow(std::shared_ptr<Flow> flow)
{
    if (!flow) {
        return;
    }

    for (std::vector<std::shared_ptr<Flow>>::iterator it = _flows.begin(); it != _flows.end(); ++it) {
        if (*it) {
            if (flow.get() == it->get()) {
                _flows.erase(it);
                break;
            }
        }
    }
}

void Link::addFlows(std::vector<std::shared_ptr<Flow> > flows)
{
    _flows.insert(_flows.end(), flows.begin(), flows.end());
}

void Link::removeAllFlows()
{
    for (std::shared_ptr<Flow> &flow : _flows) {
        flow = nullptr;
    }
    _flows.clear();
}

const std::vector<std::shared_ptr<Flow> > &Link::getAllFlows() const
{
    return _flows;
}
