/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef LINK_H
#define LINK_H

#include <memory>
#include <vector>

#include "../commonconfig.h"
#include "flow.h"

class Switcher;

/*!
 * \brief The LinkDirection enum holds the possible directions that a link connected to
 * a switcher can have.
 */
enum class LinkDirection {
    INPUT = 0,
    OUTPUT = 1
};

/*!
 * \brief The LinkLocation enum holds the possible locations that a link connected to
 * a switcher can be.
 */
enum class LinkLocation {
    LOCAL = 0,
    NORTH = 2,
    SOUTH = 4,
    EAST = 6,
    WEST = 8,
    DISCONNECTED
};

/*!
 * \brief The Link class represents a communication link in the network-on-chip.
 */
class Link
{
    /*!
     * \brief Holds a counter for the total number of links.
     */
    static unsigned _linkCounter;

    /*!
     * \brief Holds an identifiTaskcation number for this link.
     */
    unsigned _id;

    /*!
     * \brief Holds the location where this link is connected to a switcher.
     */
    LinkLocation _location;

    /*!
     * \brief Holds the list of flows that crosses this link.
     */
    std::vector<std::shared_ptr<Flow>> _flows;

    /*!
     * \brief Holds a reference to the switcher that writes to this link.
     */
    std::weak_ptr<Switcher> _sourceSwitcher;

    /*!
     * \brief Holds a reference to the switcher that reads from this link.
     */
    std::weak_ptr<Switcher> _destinationSwitcher;

public:
    /*!
     * \brief Default Constructor.
     */
    Link();

    /*!
     * \brief Copy Constructor.
     * \param other Other instance.
     */
    Link(const Link &other);

    /*!
     * \brief Default Destructor.
     */
    ~Link();

    /*!
     * \brief Checks whether a flow transverse this link.
     * \param flow A reference to a flow.
     * \return True if the flow transverse this link, False otherwise.
     */
    bool contains(const std::shared_ptr<Flow> flow);

    /*!
     * \brief Calculates the utilization rate for this link as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \return The link utilization rate.
     */
    _float calculateUtilization();

    /*!
     * \brief Adds a flow to the list of flows that transverse this link.
     * \param flow The flow that transverse this link.
     */
    void addFlow(std::shared_ptr<Flow> flow);

    /*!
     * \brief Removes a flow to the list of flows that transverse this link.
     * \param flow The flow that no longer transverse this link.
     */
    void removeFlow(std::shared_ptr<Flow> flow);

    /*!
     * \brief Adds multiple flows to the list of flows that transverse this link.
     * \param flows The flows that transverse this link.
     */
    void addFlows(std::vector<std::shared_ptr<Flow>> flows);

    /*!
     * \brief Removes all flows that transverse this link,
     */
    void removeAllFlows();

    /*!
     * \brief Gets a vector of references to all flows that crosses this link.
     * \return A vector of reference to this processor's tasks.
     */
    const std::vector<std::shared_ptr<Flow>> &getAllFlows() const;

    /*!
     * \brief Getter to this link location when connected to a switcher.
     * \return The location as \c LinkLocation.
     */
    LinkLocation getLocation();

    /*!
     * \brief Connects this link to a source switcher.
     * \param sourceSwitcher A reference to the switcher that writes into this link.
     * \param location The position that this link is connected to the switcher's router.
     */
    void connectToSourceSwitcher(std::shared_ptr<Switcher> sourceSwitcher, LinkLocation location);

    /*!
     * \brief Connects this link to a destination switcher.
     * \param destinationSwitcher A reference to the switcher that reads from this link.
     */
    void connectToDestinationSwitcher(std::shared_ptr<Switcher> destinationSwitcher);

    /*!
     * \brief Getter to the source switch for this link, i.e. the switcher that writes into it.
     * \return A reference to the source switcher.
     */
    std::weak_ptr<Switcher> getSourceSwitcher() const;

    /*!
     * \brief Getter to the destination switch for this link, i.e. the switcher that reads from it.
     * \return A reference to the destination switcher.
     */
    std::weak_ptr<Switcher> getDestinationSwitcher() const;
};

#endif // LINK_H
