﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "mapper.h"

#include <iostream>
#include <string>
#include <utility>
#include <cstdio>

Mapper::Mapper()
{   
}

Mapper::Mapper(const Mapper &other) :
    _routingElement(other._routingElement)
{
}

Mapper::~Mapper()
{
}

bool Mapper::doMapping(const std::vector<int> &tasksPlacement,
                       const std::shared_ptr<Application> app,
                       const std::shared_ptr<Platform> platform)
{
    bool returnValue = true;
    if (tasksPlacement.size() != app->getNumberOfTasks()) {
        printf("Erro - apptasks %d mtasks %d \n", app->getNumberOfTasks(),
               (int)(tasksPlacement.size()));
        throw std::runtime_error(
                    ERROR_MSG(__FILE__, __LINE__,
                              "task placement variable's size is not equal as the number "
                              "of tasks."));
    }

    // Check whether all placements are valid.
    for (int placement : tasksPlacement) {
        if (placement > static_cast<int>(Platform::getNumberOfCores() - 1) || placement < 0) {
            throw std::runtime_error(
                        ERROR_MSG(__FILE__, __LINE__,
                                  "tasks placement outside boundaries."));
        }
    }

    // Clear previous mappings
    undoMapping(platform);

    // Maps tasks to processors
    unsigned i = 0;
    for (std::shared_ptr<Task> &task : app->getTasksAsVector()) {
        std::shared_ptr<Processor> processor = platform->getProcessor(tasksPlacement[i]);
        processor->mapTask(task);
        _map[task] = processor;
        if (!(app->getNumberOfTasks())) {
            break;
        }
        i++;
    }

    // Maps flows to links   
    i = 0;
    for (std::shared_ptr<Flow> &flow : app->getFlowsAsVector()) {
        auto srcTask = flow->getSourceTask().lock();
        auto dstTask = flow->getDestinationTask().lock();
        if (srcTask && dstTask) {
            std::shared_ptr<Processor> processorSrc = _map[srcTask];
            std::shared_ptr<Processor> processorDst = _map[dstTask];
            std::vector<std::shared_ptr<Link>> links;
            returnValue &= _routingElement.doRouting(links, processorSrc, processorDst);
            for (std::shared_ptr<Link> &link : links) {
                link->addFlow(flow);
            }
            _mapFlows[flow] = links;
        } else {
            throw std::runtime_error(
                        ERROR_MSG(__FILE__, __LINE__,
                                  "flow's sourceTask or destinationTask is invalid."));
            return false;
        }

        if (!(i < app->getNumberOfFlows())) {
            break;
        }
        i++;
    }

    return returnValue;
}

void Mapper::undoMapping(const std::shared_ptr<Platform> platform)
{
    platform->clearAllTasksInProcessors();
    _map.clear();

    platform->clearAllFlowsInLinks();
    _mapFlows.clear();
}

std::shared_ptr<Processor> Mapper::mapTask(std::shared_ptr<Task> task)
{
    if (_map.size() == 0 || !task) {
        return nullptr;
    }

    if (_map.find(task) != _map.end()) {
        return _map[task];
    } else {
        return nullptr;
    }
}

const std::vector<std::shared_ptr<Link>> &Mapper::mapFlow(std::shared_ptr<Flow> flow)
{
    return _mapFlows[flow];
}

