/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef MAPPER_H
#define MAPPER_H

#include <map>
#include <vector>

#include "../commonconfig.h"
#include "application.h"
#include "flow.h"
#include "link.h"
#include "platform.h"
#include "processor.h"
#include "routingelement.h"
#include "task.h"

/*!
 * \brief The Mapper class is responsible for mapping an application to a platform.
 */
class Mapper
{
    /*!
     * \brief Holds a map of references between tasks and processors.
     */
    std::map<std::shared_ptr<Task>, std::shared_ptr<Processor>> _map;

    /*!
     * \brief Holds a map of references between flows and links.
     */
    std::map<std::shared_ptr<Flow>, std::vector<std::shared_ptr<Link>>> _mapFlows;

    /*!
     * \brief Holds an instance of a routing element to route the communication flows through
     * the platform.
     */
    RoutingElement _routingElement;

public:
    /*!
     * Default Constructor.
     */
    Mapper();

    /*!
     * \brief Copy Constructor.
     * \param other Other instance.
     */
    Mapper(const Mapper &other);

    /*!
     * \brief Defaut Destructor.
     */
    ~Mapper();

    /*!
     * \brief Executes the mapping of an application and its tasks and flows to a platform.
     * \param A vector of integers holding the placement of tasks to processors.
     * \param platform A reference to the system's platform.
     * \param app A reference to the application being mapped.
     * \param platform A reference to the platform.
     * \return true if the mapping succed, false otherwise.
     */
    bool doMapping(const std::vector<int> &tasksPlacement,
                   const std::shared_ptr<Application> app,
                   const std::shared_ptr<Platform> platform);


    /*!
     * \brief Undo the application to platform mapping, i.e. resets the platform as there is no
     * application being executed.
     * \param A reference to the platform being unmapped.
     */
    void undoMapping(const std::shared_ptr<Platform> platform);


    /*!
     * \brief Maps a task to its processor as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \param task A reference to a task.
     * \return A reference to the processor that executes this task. If the tasks were not mapped yet
     * this method returns \c nullptr.
     */
    std::shared_ptr<Processor> mapTask(std::shared_ptr<Task> task);

    /*!
     * \brief Maps a flow to its links as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \param flow A reference to a flow.
     * \return An array of links that this flow crosses during communication.
     */
    const std::vector<std::shared_ptr<Link> > &mapFlow(std::shared_ptr<Flow> flow);

};

#endif // MAPPER_H
