﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "platform.h"

unsigned Platform::_numberOfRows = 4;

unsigned Platform::_numberOfColumns = 4;

unsigned Platform::_numberOfCores = 16;

unsigned Platform::_flitSizeInBits = 32;

unsigned Platform::_numberOfLinks = 80;

_float Platform::_latencyLinkHop = 2e-8;

_float Platform::_latencyRouterHop = 2e-7;

unsigned Platform::getNumberOfRows()
{
    return _numberOfRows;
}

unsigned Platform::getNumberOfColumns()
{
    return _numberOfColumns;
}

unsigned Platform::getNumberOfCores()
{
    return _numberOfCores;
}

unsigned Platform::getNumberOfLinks()
{
    return _numberOfLinks;
}

unsigned Platform::getFlitSizeInBits()
{
    return _flitSizeInBits;
}


_float Platform::getLatencyLinkHop()
{
    return _latencyLinkHop;
}

_float Platform::getLatencyRouterHop()
{
    return _latencyRouterHop;
}

std::shared_ptr<Processor> Platform::getProcessor(int processorIndex)
{
    return _processors.at(processorIndex);
}

std::shared_ptr<Link> Platform::getLink(int linkIndex)
{
    return _links.at(linkIndex);
}

Platform::Platform()
{
    _createNocElements();
}

Platform::Platform(const Platform &other)
{
    for (const auto& proc : other._processors) {
        _processors.push_back(std::make_shared<Processor>(*proc));
    }

    for (const auto& link : other._links) {
        _links.push_back(std::make_shared<Link>(*link));
    }

    for (const auto& switcher : other._switchers) {
        _switchers.push_back(std::make_shared<Switcher>(*switcher));
    }
}

Platform::~Platform()
{
    _clearNocElements();
}

void Platform::_createNocElements()
{
    // Clear platform previous elements.
    _clearNocElements();

    // Create Processors, Switchers and Local Links
    for (unsigned i = 0; i < _numberOfCores; i++) {
        std::shared_ptr<Processor> processor = std::make_shared<Processor>();
        _processors.push_back(processor);
        std::shared_ptr<Switcher> switcher = std::make_shared<Switcher>();
        switcher->setAddress(i);
        processor->connectToSwitcher(switcher);
        _switchers.push_back(switcher);
        std::shared_ptr<Link> linkIn = std::make_shared<Link>();
        std::shared_ptr<Link> linkOut = std::make_shared<Link>();
        _links.push_back(linkIn);
        _links.push_back(linkOut);
        switcher->connectLinksLocal(linkIn, linkOut);
        linkIn->connectToDestinationSwitcher(switcher);
        linkOut->connectToSourceSwitcher(switcher, LinkLocation::LOCAL);
    }

    // Create the other links
    for (unsigned i = 0; i < _numberOfCores; i++) {
        // Connect in Same Row
        if ((i / _numberOfRows) == ((i + 1) / _numberOfRows)) {
            std::shared_ptr<Link> linkIn = std::make_shared<Link>();
            std::shared_ptr<Link> linkOut = std::make_shared<Link>();
            _links.push_back(linkIn);
            _links.push_back(linkOut);
            _switchers[i]->connectLinksEast(linkIn, linkOut);
            linkIn->connectToDestinationSwitcher(_switchers[i]);
            linkOut->connectToSourceSwitcher(_switchers[i], LinkLocation::EAST);
            _switchers[i + 1]->connectLinksWest(linkOut, linkIn);
            linkOut->connectToDestinationSwitcher(_switchers[i + 1]);
            linkIn->connectToSourceSwitcher(_switchers[i + 1], LinkLocation::WEST);
        }

        // Connect in Same Column
        if (_numberOfCores != _numberOfRows && i >= _numberOfRows) {
            std::shared_ptr<Link> linkIn = std::make_shared<Link>();
            std::shared_ptr<Link> linkOut = std::make_shared<Link>();
            _links.push_back(linkIn);
            _links.push_back(linkOut);
            _switchers[i - _numberOfRows]->connectLinksSouth(linkIn, linkOut);
            linkIn->connectToDestinationSwitcher(_switchers[i - _numberOfRows]);
            linkOut->connectToSourceSwitcher(_switchers[i - _numberOfRows], LinkLocation::SOUTH);
            _switchers[i]->connectLinksNorth(linkOut, linkIn);
            linkOut->connectToDestinationSwitcher(_switchers[i]);
            linkIn->connectToSourceSwitcher(_switchers[i], LinkLocation::NORTH);
        }
    }
}

void Platform::_clearNocElements()
{
    for (auto &processor : _processors) {
        processor = nullptr;
    }
    _processors.clear();

    for (auto &link : _links) {
        link = nullptr;
    }
    _links.clear();

    for (auto &switcher : _switchers) {
        switcher = nullptr;
    }
    _switchers.clear();
}

void Platform::reconfigureNoc(unsigned numberOfRows, unsigned numberOfCols)
{
    _processors.clear();
    _numberOfRows = numberOfRows;
    _numberOfColumns = numberOfCols;
    _numberOfCores = _numberOfRows * _numberOfColumns;
    _numberOfLinks = (_numberOfCores * 2) + ((_numberOfColumns - 1) * 2) * _numberOfRows +
            ((_numberOfRows - 1) * 2) * _numberOfColumns;
    _createNocElements();
}

void Platform::reconfigureLatencies(_float latencyLink, _float latencyRouter)
{
    _latencyLinkHop = latencyLink;
    _latencyRouterHop = latencyRouter;
}

void Platform::reconfigureFlits(unsigned flitSizeInbits)
{
    _flitSizeInBits = flitSizeInbits;
}

void Platform::clearAllTasksInProcessors()
{
    for (auto &processor : _processors) {
        processor->unmapAllTasks();
    }
}

void Platform::clearAllFlowsInLinks()
{
    for (auto &link : _links) {
        link->removeAllFlows();
    }
}
