/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef PLATFORM_H
#define PLATFORM_H

#include <map>
#include <memory>
#include <vector>

#include "../commonconfig.h"

#include "link.h"
#include "processor.h"
#include "switcher.h"

/*!
 * \brief The \c Platform class represents the system-on-a-chip platform that uses a Priority-based
 * Wormhole RT-NOC. The system has as its elements: processors, links and switches (routers).
 * \sa \c Processor, \c Link, \c Switcher
 */
class Platform
{
    /*!
     * \brief Holds the number of rows in the mesh-grid network-on-chip.
     * \note Default value is 4.
     */
    static unsigned _numberOfRows;

    /*!
     * \brief Holds the number of columns in the mesh-grid network-on-chip.
     * \note Default value is 4.
     */
    static unsigned _numberOfColumns;

    /*!
     * \brief Holds the number of cores in the system-on-chip.
     * \note Always holds the value of \c _numberOfColumns times \c _numberOfRows.
     * \note Default value is 16.
     */
    static unsigned _numberOfCores;

    /*!
     * \brief Holds the number of links in the system-on-chip.
     * \note Default value is 80.
     */
    static unsigned _numberOfLinks;

    /*!
     * \brief Holds the size of a flit in bits.
     * \note Default value is 32 bits.
     */
    static unsigned _flitSizeInBits;

    /*!
     * \brief Holds the delay caused when a flit pass through a link.
     * \note Default value is 2e-8 seconds.
     */
    static _float _latencyLinkHop;

    /*!
     * \brief Holds the delay caused when a flit pass through a router.
     * \note Default value is 2e-7 seconds.
     */
    static _float _latencyRouterHop;

    /*!
     * \brief Holds the processors in this platform.
     */
    std::vector<std::shared_ptr<Processor>> _processors;

    /*!
     * \brief Holds the communication links in this platform.
     */
    std::vector<std::shared_ptr<Link>> _links;

    /*!
     * \brief Holds the switchers in this platform.
     */
    std::vector<std::shared_ptr<Switcher>> _switchers;

    /*!
     * \brief Creates the components of this platform.
     */
    void _createNocElements();

    /*!
     * \brief Clears the components of this platform.
     */
    void _clearNocElements();

public:
    /*!
     * \brief Default constructor.
     */
    Platform();

    /*!
     * \brief Copy Constructor.
     * \param other Other instance of platform.
     */
    Platform(const Platform &other);

    /*!
     * \brief Default Constructor.
     */
    ~Platform();

    /*!
     * \brief Reconfigure platform cores parameters.
     */
    void reconfigureNoc(unsigned numberOfRows, unsigned numberOfCols);

    /*!
     * \brief Reconfigure platform communication latencies.
     */
    void reconfigureLatencies(_float latencyLink, _float latencyRouter);

    /*!
     * \brief reconfigureFlits
     */
    void reconfigureFlits(unsigned flitSizeInbits);

    /*!
     * \brief Clears all tasks mapped to the platform's processors.
     */
    void clearAllTasksInProcessors();

    /*!
     * \brief Clears all flows that crosess the platform's links.
     */
    void clearAllFlowsInLinks();

    /*!
     * \brief Gets a reference for one of the platform's processors.
     * \param processorIndex The processor index.
     * \return The processor reference.
     */
    std::shared_ptr<Processor> getProcessor(int processorIndex);

    /*!
     * \brief Gets a reference for one of the platform's links.
     * \param linkIndex The link index.
     * \return The link reference.
     */
    std::shared_ptr<Link> getLink(int linkIndex);

    /*!
     * \brief Getter to number of rows in this platform.
     */
    static unsigned getNumberOfRows();

    /*!
     * \brief Getter to number of columns in this platform.
     */
    static unsigned getNumberOfColumns();

    /*!
     * \brief Getter to number of cores in this platform.
     */
    static unsigned getNumberOfCores();

    /*!
     * \brief Getter to number of links in this platform.
     */
    static unsigned getNumberOfLinks();

    /*!
     * \brief Getter to this platform flits' size in bits.
     */
    static unsigned getFlitSizeInBits();

    /*!
     * \brief Getter to this platform hop delay in links in seconds.
     */
    static _float getLatencyLinkHop();

    /*!
     * \brief Getter to this platform hop delay in routers in seconds.
     */
    static _float getLatencyRouterHop();
};

#endif // PLATFORM_H
