﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "processor.h"

#include <iostream>

unsigned Processor::_processorCounter = 0;

std::shared_ptr<Switcher> Processor::getSwitcher() const
{
    return _switcher;
}

const std::vector<std::shared_ptr<Task> > &Processor::getAllMappedTasks() const
{
    return _tasks;
}

Processor::Processor() :
    _id(++_processorCounter),
    _switcher(nullptr)
{
}

Processor::Processor(const Processor &other) :
    _id(++_processorCounter),
    _switcher(std::make_shared<Switcher>(*other._switcher))
{
    for (const auto &task : other._tasks) {
        _tasks.push_back(std::make_shared<Task>(*task));
    }
}

Processor::~Processor()
{
    --_processorCounter;

    for (auto &task : _tasks) {
        task = nullptr;
    }

    _switcher = nullptr;
}

bool Processor::contains(const std::shared_ptr<Task> task)
{
    bool ret = false;
    for (const std::shared_ptr<Task> mappedTask : _tasks) {
        if (mappedTask == task) {
            ret = true;
            break;
        }
    }
    return ret;
}

_float Processor::calculateUtilization()
{
    _float utilization = 0.0;
    for (std::shared_ptr<Task> task : _tasks) {
        if (task) {
            utilization += task->getCost() / task->getPeriod();
        }
    }
    return utilization;
}

void Processor::mapTask(std::shared_ptr<Task> task)
{
    _tasks.push_back(task);
}

void Processor::unmapTask(std::shared_ptr<Task> task)
{
    if (!task) {
        return;
    }

    for (std::vector<std::shared_ptr<Task>>::iterator it = _tasks.begin(); it != _tasks.end(); ++it) {
        if (*it) {
            if (task.get() == it->get()) {
                _tasks.erase(it);
                break;
            }
        }
    }
}

void Processor::mapTasks(std::vector<std::shared_ptr<Task> > tasks)
{
    _tasks.insert(_tasks.end(), tasks.begin(), tasks.end());
}

void Processor::unmapAllTasks()
{
    for (std::shared_ptr<Task> &task : _tasks) {
        task = nullptr;
    }
    _tasks.clear();
}

void Processor::connectToSwitcher(std::shared_ptr<Switcher> switcher)
{
    _switcher = switcher;
}
