/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef PROCESSOR_H
#define PROCESSOR_H

#include <memory>
#include <vector>

#include "../commonconfig.h"
#include "task.h"
#include "switcher.h"

/*!
 * \brief The Processor class represents a processing core in the
 * system-on-chip platform.
 * As described in:
 * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
 */
class Processor
{
    /*!
     * \brief Holds a counter to the total of processors.
     */
    static unsigned _processorCounter;

    /*!
     * \brief Holds an id number for this processor.
     */
    unsigned _id;

    /*!
     * \brief Holds the list of tasks that this processor will execute.
     */
    std::vector<std::shared_ptr<Task>> _tasks;

    /*!
     * \brief Holds a reference to the switcher element connected to this processor.
     */
    std::shared_ptr<Switcher> _switcher;

public:
    /*!
     * \brief Default Constructor.
     */
    Processor();

    /*!
     * \brief Copy Constructor.
     * \param other The other instance.
     */
    Processor(const Processor& other);

    /*!
     * \brief Default Destructor.
     */
    ~Processor();

    /*!
     * \brief Checks whether a task was mapped to this processor.
     * \param task A reference to a task.
     * \return True if this processor has this task mapped,
     * False otherwise.
     */
    bool contains(const std::shared_ptr<Task> task);

    /*!
     * \brief Calculates the utilization rate for this processor.
     * The result is also stored in the processor utilization rate.
     * This algorithm behaves as described in:
     * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \return This processor utilization rate.
     */
    _float calculateUtilization();

    /*!
     * \brief Maps a task to this processor.
     * \param task The task being mapped.
     */
    void mapTask(std::shared_ptr<Task> task);

    /*!
     * \brief Unmaps a task to this processor.
     * \param task the task being unmapped.
     */
    void unmapTask(std::shared_ptr<Task> task);

    /*!
     * \brief Maps multiple tasks to this processor.
     * \param tasks The tasks being mapped.
     */
    void mapTasks(std::vector<std::shared_ptr<Task>> tasks);

    /*!
     * \brief Unmaps all tasks in this processor.
     */
    void unmapAllTasks();

    /*!
     * \brief Connects this processor to a switcher.
     * \param switcher A reference to a switcher.
     */
    void connectToSwitcher(std::shared_ptr<Switcher> getSwitcher);

    /*!
     * \brief Getter to the switcher connected to this processor.
     * \return A reference to this processor's switcher.
     */
    std::shared_ptr<Switcher> getSwitcher() const;

    /*!
     * \brief Gets a vector of references to all tasks mapped to this processor.
     * \return A vector of reference to this processor's tasks.
     */
    const std::vector<std::shared_ptr<Task>> &getAllMappedTasks() const;
};

#endif // PROCESSOR_H
