﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "routingelement.h"

#include <cmath>
#include <iostream>

#include "platform.h"

RoutingElement::RoutingElement()
{

}

RoutingElement::~RoutingElement()
{

}

bool RoutingElement::doRouting(std::vector<std::shared_ptr<Link>> &links,
                               const std::shared_ptr<Processor> &processorSource,
                               const std::shared_ptr<Processor> &processorDestination)
{
    if (!processorSource) {
        throw std::runtime_error(
                    std::string("Error: RoutingElement::doRouting:") +
                    std::to_string(__LINE__) +
                    std::string(" routing processor source reference is invalid."));
    }

    if (!processorDestination) {
        throw std::runtime_error(
                    std::string("Error: RoutingElement::doRouting:") +
                    std::to_string(__LINE__) +
                    std::string(" routing processor destination reference is invalid."));
    }

    if (!processorSource->getSwitcher()) {
        throw std::runtime_error(
                    std::string("Error: RoutingElement::doRouting:") +
                    std::to_string(__LINE__) +
                    std::string(" routing processor source is not connected to a switcher."));
    }

    if (!processorDestination->getSwitcher()) {
        throw std::runtime_error(
                    std::string("Error: RoutingElement::doRouting:") +
                    std::to_string(__LINE__) +
                    std::string(" routing processor destination is not connected to a switcher."));
    }

    unsigned srcAddress = processorSource->getSwitcher()->getAddress();
    unsigned dstAddress = processorDestination->getSwitcher()->getAddress();

    // Make sure the list is empty
    links.clear();

    // Check whether both tasks are mapped to the same processor
    if (srcAddress == dstAddress) {
        return true;
    }

    unsigned rowSrc = srcAddress / Platform::getNumberOfRows();
    unsigned rowDst = dstAddress / Platform::getNumberOfRows();

    unsigned colSrc = srcAddress % Platform::getNumberOfRows();
    unsigned colDst = dstAddress % Platform::getNumberOfRows();

    int deltaRows = int(rowDst) - int(rowSrc);
    int deltaCols = int(colDst) - int(colSrc);
    int absDeltaRows = std::abs(deltaRows);
    int absDeltaCols = std::abs(deltaCols);
    bool signDeltaRows = deltaRows >= 0;
    bool signDeltaCols = deltaCols >= 0;

    // Adds the source local link
    auto link = processorSource->getSwitcher()->getLinksLocal().first;
    links.push_back(link);

    // Moves in Row
    for (int i = 0; i < absDeltaCols; i++) {
        auto swicher = link->getDestinationSwitcher().lock();
        if (!swicher) {
            throw std::runtime_error(std::string("Error: RoutingElement::doRouting:") +
                                     std::to_string(__LINE__) +
                                     std::string(" routing through rows found a invalid switcher reference."));
        }

        link = signDeltaCols ? swicher->getLinksEast().second : swicher->getLinksWest().second;
        links.push_back(link);
    }

    // Moves in Column
    for (int i = 0; i < absDeltaRows; i++) {
        auto swicher = link->getDestinationSwitcher().lock();
        if (!swicher) {
            throw std::runtime_error(std::string("Error: RoutingElement::doRouting:") +
                                     std::to_string(__LINE__) +
                                     std::string(" routing through columns found a invalid switcher reference."));
        }

        link = signDeltaRows ? swicher->getLinksSouth().second : swicher->getLinksNorth().second;
        links.push_back(link);
    }

    // Checks whether the routing was successfull
    auto swicher = link->getDestinationSwitcher().lock();
    if (!swicher) {
        throw std::runtime_error(std::string("Error: RoutingElement::doRouting:") +
                                 std::to_string(__LINE__) +
                                 std::string(" routing destination has a invalid swithcer reference."));
    }

    if (swicher->getAddress() != dstAddress) {
        links.clear();
        std::cerr << "Error: Routing was not successfull."
                  << std::endl;
        return false;
    }

    // Adds the destination local link
    link = swicher->getLinksLocal().second;
    links.push_back(link);

    return true;
}
