﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "switcher.h"

unsigned Switcher::_switcherCounter = 0;

Switcher::Switcher() :
    _id(++_switcherCounter),
    _address(0)
{
    for (int i = 0; i < 10; i++) {
        _links.push_back(nullptr);
    }
}

Switcher::Switcher(const Switcher &other) :
    _id(++_switcherCounter),
    _address(other._address)
{
    for (const auto &link : other._links) {
        if (link) {
            _links.push_back(std::make_shared<Link>(*link));
        } else {
            _links.push_back(nullptr);
        }
    }
}

unsigned Switcher::getAddress()
{
    return _address;
}

void Switcher::setAddress(const unsigned &address)
{
    _address = address;
}

void Switcher::_connectLinks(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink, LinkLocation location)
{
    _links[int(LinkDirection::INPUT) + int(location)] = inputLink;
    _links[int(LinkDirection::OUTPUT) + int(location)] = outputLink;
}

Switcher::~Switcher()
{
    _switcherCounter--;

    for (auto &link : _links) {
        link = nullptr;
    }
    _links.clear();
}

void Switcher::connectLinksLocal(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink)
{
    _connectLinks(inputLink, outputLink, LinkLocation::LOCAL);
}

void Switcher::connectLinksNorth(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink)
{
    _connectLinks(inputLink, outputLink, LinkLocation::NORTH);
}

void Switcher::connectLinksSouth(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink)
{
    _connectLinks(inputLink, outputLink, LinkLocation::SOUTH);
}

void Switcher::connectLinksEast(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink)
{
    _connectLinks(inputLink, outputLink, LinkLocation::EAST);
}

void Switcher::connectLinksWest(std::shared_ptr<Link> inputLink, std::shared_ptr<Link> outputLink)
{
    _connectLinks(inputLink, outputLink, LinkLocation::WEST);
}

std::pair<std::shared_ptr<Link>, std::shared_ptr<Link> > Switcher::getLinksLocal()
{
    return std::make_pair(_links[int(LinkDirection::INPUT) + int(LinkLocation::LOCAL)],
            _links[int(LinkDirection::OUTPUT) + int(LinkLocation::LOCAL)]);
}

std::pair<std::shared_ptr<Link>, std::shared_ptr<Link> > Switcher::getLinksNorth()
{
    return std::make_pair(_links[int(LinkDirection::INPUT) + int(LinkLocation::NORTH)],
            _links[int(LinkDirection::OUTPUT) + int(LinkLocation::NORTH)]);
}

std::pair<std::shared_ptr<Link>, std::shared_ptr<Link> > Switcher::getLinksSouth()
{
    return std::make_pair(_links[int(LinkDirection::INPUT) + int(LinkLocation::SOUTH)],
            _links[int(LinkDirection::OUTPUT) + int(LinkLocation::SOUTH)]);
}

std::pair<std::shared_ptr<Link>, std::shared_ptr<Link> > Switcher::getLinksEast()
{
    return std::make_pair(_links[int(LinkDirection::INPUT) + int(LinkLocation::EAST)],
            _links[int(LinkDirection::OUTPUT) + int(LinkLocation::EAST)]);
}

std::pair<std::shared_ptr<Link>, std::shared_ptr<Link> > Switcher::getLinksWest()
{
    return std::make_pair(_links[int(LinkDirection::INPUT) + int(LinkLocation::WEST)],
            _links[int(LinkDirection::OUTPUT) + int(LinkLocation::WEST)]);
}

void Switcher::disconnectAllLinks()
{
    _links.clear();
}

bool Switcher::contains(std::shared_ptr<Link> link)
{
    bool ret = false;
    if (link) {
        for (const std::shared_ptr<Link> connectedLink : _links) {
            if (connectedLink == link) {
                ret = true;
                break;
            }
        }
    }
    return ret;
}


