/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef SYSTEM_H
#define SYSTEM_H

#include <array>
#include <map>
#include <memory>
#include <vector>

#include "../commonconfig.h"
#include "application.h"
#include "mapper.h"
#include "platform.h"
#include "task.h"

/*!
 * \brief The System class represents the top-level system, i.e. both the SoC platform and the
 * real-time application.
 * \sa Platform, Application.
 */
class System
{
    /*!
     * \brief Holds an array with the values of all possible frequency factors for the system.
     * The values for this were generated getting 255 points in a logarithmic
     * space between 0.01 and 100 using the function logspace(-2, 2, 255) in Matlab.
     */
    static const std::array<_float, 255> _frequencyFactors;

    /*!
     * \brief Holds a reference to the platform.
     */
    std::shared_ptr<Platform> _platform;

    /*!
     * \brief Holds a reference to the application.
     */
    std::shared_ptr<Application> _application;

    /*!
     * \brief Holds whether there is an application mapped to the platform.
     */
    bool _isMapped;

    /*!
     * \brief Holds an instance of a mapper to map the application to the platform.
     */
    Mapper _mapper;

    /*!
     * \brief Calculates platform's flows worst case latencies as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \return true if the calculation is successfull, false otherwise.
     */
    bool _calculateWorstCaseLatencies();

    /*!
     * \brief Calculates platform's flows worst case latencies as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * However, in this method, the system frequency is also considered to calculate the worst case
     * latency as described in:
     * \return true if the calculation is successfull, false otherwise.
     */
    bool _calculateScaledWorstCaseLatencies(_float frequencyFactor);

    /*!
     * \brief Check whether the system with the current task mapping placement is completely
     * schedulable at a specific frequency factor.
     * \param frequency A frequency factor for the system.
     * \return true if the system is completely schedulable running with a specific frequency,
     * false. otherwise.
     */
    bool _schedulabilityTest(_float frequencyFactor);

    /*!
     * \brief Calculates the system breakdown frequency, i.e. the frequency factor in which all the
     * application tasks are schedulable.
     * \return The breakdown frequency.
     */
    _float _calculateBreakdownFrequency();

public:
    /*!
     * \brief Default Constructor initializes the system with an application.
     * \param app The application processed by this system.
     */
    System(std::shared_ptr<Application> app);

    /*!
     * \brief Copy constructor,
     * \param other Other instance of system.
     */
    System(const System &other);

    /*!
     * \brief Default Constructor.
     */
    ~System();

    /*!
     * \brief Reconfigures the system's platform.
     */
    void reconfigurePlatform(unsigned numberOfRows, unsigned numberOfCols);

    /*!
     * \brief Reconfigures the system's application.
     */
    void reconfigureApplication(const std::string& fileName);

    /*!
     * \brief Get application's name.
     */
    std::string getApplicationName() const;

    /*!
     * \brief Getter to the number of tasks in the application.
     */
    unsigned getNumberOfApplicationTasks();

    /*!
     * \brief Getter to the number of processors in the platform.
     */
    unsigned getNumberOfPlatformProcessors();

    /*!
     * \brief Getter to the number of processor rows in the Mesh-Grid Platform.
     */
    unsigned getNumberOfProcessorRows();

    /*!
     * \brief Getter to the number of processor columns in the Mesh-Grid Platform.
     */
    unsigned getNumberOfProcessorColumns();

    /*!
     * \brief Check whether there is an application mapped in the system's platform.
     * \return True if there is, False otherwise.
     */
    bool isMapped() const;

    /*!
     * \brief Performs a mapping of the application to the platform given a list of values for the
     * processors positions for the application's tasks.
     * \return true if the mapping succed, false otherwise.
     */
    bool doMapping(const std::vector<int> &taskPlacement);

    /*!
     * \brief Calculates all platform's processors utilization rates.
     * \param utilizationRates A vector that will contain the results of utilization rates.
     * \return true if the calculation is successful, false otherwise.
     */
    bool calculateProcessorsUtilizationRate(std::vector<_float> &utilizationRates);

    /*!
     * \brief Calculates all platform's flows basic latency.
     * \param basicLatencies A vector that will be populated with the values of the system's
     * basic latencies.
     * \return true if the calculation is successful, false otherwise.
     */
    bool calculateFlowsBasicLatency(std::vector<_float> &basicLatencies);

    /*!
     * \brief Calculates all platform's tasks response time.
     * \param responseTimes A vector that will be populated with the response times.
     * \return true if the calculation is successful, false otherwise.
     */
    bool calculateTasksResponseTime(std::vector<_float> &responseTimes);

    /*!
     * \brief Calculates all platform's flows worst case latencies.
     * \param worstCaseLatencies A vector that will be populated with the values of the worst
     * case latencies.
     * \note This method uses: \c calculateTasksResponseTime, \c calculateFlowsBasicLatency.
     * \return true if the calculation is successful, false otherwise.
     */
    bool calculateFlowsWorstCaseLatency(std::vector<_float> &worstCaseLatencies);

    /*!
     * \brief Calculates platform's tasks end-to-end response times as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \param endToEndResponses A vector that will be populated with the values of the end-to-end response time.
     * \return true if the calculation is successfull, false otherwise.
     */
    bool calculateTasksEndToEndResponseTime(std::vector<_float> &endToEndResponses);

    /*!
     * \brief Calculates platform's tasks end-to-end response times as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * However, for this method the system's frequency factor is also considered.
     * \param frequencyFactor A factor that indicates how fast is the system's frequency a parameter
     * that influences the response time and latencies for the application tasks.
     * \param endToEndResponses A vector that will be populated with the values of the end-to-end response time.
     * \return true if the calculation is successfull, false otherwise.
     */
    bool calculateScaledTasksEndToEndResponseTime(_float frequencyFactor, std::vector<_float> &endToEndResponses);

    /*!
     * \brief Performs the End-to-End Schedulability Analysis for all system's tasks as described in:
     * \sa Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
     * \param tasksSchedulability A vector of \c bool that are populated representing a list of unschedulable tasks.
     * \note If this method is used and the system was not mapped before it will return false.
     * \return true if the analysis is successfull, false otherwise.
     */
    bool doEndToEndSchedulabityAnalysis(std::vector<bool> &tasksSchedulability);

    /*!
     * \brief Performs an utilization analysis for the system's processors checking whether for
     * this specific mapping configuration they are running above their capabilities.
     * \param processorsSchedulability A vector of \c bool that are populated representing a list
     * of processors above their capabilities.
     * \note If this method is used and the system was not mapped before it will return false.
     * \return true if the analysis is successfull, false otherwise.
     */
    bool doProcessorsUtilizationAnalysis(std::vector<bool> &processorsSchedulability);

    /*!
     * \brief Performs an utilization analysis for the system's links checking whether for
     * this specific mapping configuration they are running above their capabilities.
     * \param linksSchedulability A vector of \c bool that are populated representing a list
     * of links above their transmission capabilities.
     * \return true if the analysis is sucessfull, false otherwise.
     */
    bool doLinksUtilizationAnalysis(std::vector<bool> &linksSchedulability);

    /*!
     * \brief Performs a breakdown frequency analysis to identify the platform's breakdown frequency
     * in which the system is schedulable as described in:
     * \sa Sayuti, M. N. S. M., & Indrusiak, L. S. (2015). A Function for Hard Real-Time System Search-Based Task Mapping Optimisation. International Symposium on Real-Time Distributed Computing, 18, 66–73. https://doi.org/10.1109/ISORC.2015.37
     * \param breakdownFrequency A float value reference that wil be populated with frequency scale.
     * \return true if the analysis is successfull, false otherwise.
     */
    bool doBreakDownFrequencyAnalysis(_float &breakdownFrequency);

    /*!
     * \brief Performs a largest processing time first schedulling algorithm to obtain the tasks
     * placements in the platform processors.
     * \param initialPlacement Adds an initial placement for some tasks.
     * \note In this placement the tasks the placement processors are indexed from 1 to N,
     * instead of 0 to N-1. Also, elements containing 0 indicates that that task is not mapped to a processor yet.
     * \return Returns a vector of the position placement for every task in a processor index.
     * \note The placement obtained by this method can not be directly mapped to the system's (\sa doMapping).
     * Since system's platform processors are indexed from 0 to N-1 processors and this placement
     * ranges from 1 to N processors.
     */
    std::vector<int> getPlacementsLargestProcessingTimeFirst(std::vector<int> initialPlacement);

    /*!
     * \brief Performs a largest utilization first schedulling algorithm to obtain the tasks placements
     * in the platform processors.
     * \param initialPlacement Adds an initial placement for some tasks.
     * \note In this placement the tasks the placement processors are indexed from 1 to N,
     * instead of 0 to N-1. Also, elements containing 0 indicates that that task is not mapped to a processor yet.
     * \return Returns a vector of the position placement for every task in a processor index.
     * \note The placement obtained by this method can not be directly mapped to the system's (\sa doMapping).
     * Since system's platform processors are indexed from 0 to N-1 processors and this placement
     * ranges from 1 to N processors.
     */
    std::vector<int> getPlacementsLargestUtilizationFirst(std::vector<int> initialPlacement);
};

#endif // SYSTEM_H
