﻿/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "task.h"

#include <cmath>
#include <limits>
#include <stdint.h>
#include <string>

unsigned Task::_taskCounter = 0;

_float Task::_calculateResponseTime(_float periodFactor, std::vector<std::shared_ptr<Task> > &higherPriorityTasks)
{
    float i_n = 0.0;
    _float r_n = 0.0;
    _float r_n_1 = getCost() * periodFactor;

    while (true) {
        // Updates
        r_n = r_n_1;

        // Interference
        i_n = 0.0;
        for (std::shared_ptr<Task> task_j : higherPriorityTasks) {
            i_n += std::ceil(r_n / task_j->getPeriod()) *
                    task_j->getCost();
        }

        // New RTA
        r_n_1 = (getCost() + i_n) * periodFactor;

        // Check to exit loop
        if ((r_n_1 - r_n) > TIME_EPSILON && r_n_1 <= getDeadline()) {
            continue;
        } else {
            break;
        }
    }

    _responseTime = r_n_1 > getDeadline() ? UNIVERSE_AGE_IN_SEC : r_n_1;

    return _responseTime;
}

Task::Task() :
    _id(++_taskCounter),
    _cost(0.0),
    _deadline(std::numeric_limits<_float>::max()),
    _period(std::numeric_limits<_float>::max()),
    _priority(std::numeric_limits<unsigned>::max()),
    _offsetJitter(0.0),
    _flow(nullptr),
    _responseTime(-1.0)
{
}

Task::~Task()
{
     _taskCounter--;
     _flow = nullptr;
}

Task::Task(std::string name, std::string chainName, _float cost, _float deadline,
           _float period, unsigned priority, const std::shared_ptr<Flow> &flow) :
    _id(++_taskCounter),
    _name(name),
    _chainName(chainName),
    _cost(cost),
    _deadline(deadline),
    _period(period),
    _priority(priority),
    _offsetJitter(0.0),
    _flow(flow),
    _responseTime(-1.0)
{
}

Task::Task(const Task &other) :
    _id(++_taskCounter),
    _name(other._name),
    _chainName(other._chainName),
    _cost(other._cost),
    _deadline(other._deadline),
    _period(other._period),
    _priority(other._priority),
    _offsetJitter(other._offsetJitter),
    _flow(nullptr),
    _responseTime(other._responseTime)
{
    if (other._flow) {
        std::make_shared<Flow>(*other._flow);
    }
}

unsigned int Task::getUniqueId()
{
    return _id;
}

void Task::setPriority(unsigned priority)
{
    _priority = priority;
}

unsigned Task::getPriority()
{
    return _priority;
}

void Task::setName(const std::string name)
{
    _name = name;
}

const std::string Task::getName()
{
    return _name;
}

void Task::setChainName(const std::string chainName)
{
    _chainName = chainName;
}

const std::string Task::getChainName()
{
    return _chainName;
}

void Task::setCost(_float cost)
{
    _cost = cost;
}

_float Task::getCost()
{
    return _cost;
}

void Task::setDeadline(_float deadline)
{
    _deadline = deadline;
}

_float Task::getDeadline()
{
    return _deadline;
}

void Task::setPeriod(_float period)
{
    _period = period;
}

_float Task::getPeriod()
{
    return _period;
}

void Task::setOffsetJitter(_float offsetJitter)
{
    _offsetJitter = offsetJitter;
}

_float Task::getOffsetJitter()
{
    return _offsetJitter;
}

void Task::setFlow(const std::shared_ptr<Flow> &flow)
{
    _flow = flow;
}

std::shared_ptr<Flow> Task::getFlow() const
{
    return _flow;
}

_float Task::getResponseTime()
{
    return _responseTime;
}

_float Task::calculateResponseTime(const std::vector<std::shared_ptr<Task>> &sameProcTasks)
{
    // Creates the subset of tasks with higher priorities.
    std::vector<std::shared_ptr<Task>> hpTasks;
    for (auto task : sameProcTasks) {
        if (task->getPriority() <= getPriority() && task.get() != this) {
            hpTasks.push_back(task);
        }
    }

    // Calculates the RTA
    return _calculateResponseTime(1.0, hpTasks);

}

_float Task::calculateScaledResponseTime(_float frequencyFactor,
                                         const std::vector<std::shared_ptr<Task>> &sameProcTasks)
{
    // Creates the subset of tasks with higher priorities.
    std::vector<std::shared_ptr<Task>> hpTasks;
    for (auto task : sameProcTasks) {
        if (task->getPriority() <= getPriority() && task.get() != this) {
            hpTasks.push_back(task);
        }
    }

    _float periodFactor = 1 / frequencyFactor;

    // Calculates the RTA
    return _calculateResponseTime(periodFactor, hpTasks);
}
