/*!
 * \author Jessé Barreto de Barros
 * Universidade de Brasília - UnB \n
 * Departamento de Engenharia Mecânica - ENM \n
 * Laboratory of Embedded Systems and Integrated Circuits Applications - LEIA \n
 * \copyright 
 * BraSBOMF - Brasília's Search-Based Optimization Meta-heuristic Framework
 * Copyright (C) 2017-2019 Laboratory of Embedded Systems and Integrated Circuits Applications (LEIA) Universidade de Brasília
 * Jessé B. de Barros (jesseh.barreto@gmail.com)
 * Carlos H. Llanos Q. (llanos@unb.br)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef TASK_H
#define TASK_H

#include <string>
#include <memory>

#include "../commonconfig.h"
#include "flow.h"

/*!
 * \brief The Task class represents a task that can release jobs to be processed in a specific
 * processor  in the system. Implemented based in the descriptions in:
 * Indrusiak, L. S. (2014). End-to-end schedulability tests for multiprocessor embedded systems based on networks-on-chip with priority-preemptive arbitration. Journal of Systems Architecture, 60(7), 553–561. https://doi.org/10.1016/j.sysarc.2014.05.002
 */
class Task
{
    /*!
     * \brief Holds a counter to total number of tasks.
     */
    static unsigned int _taskCounter;

    /*!
     * \brief Holds the task's unique id number.
     */
    const unsigned int _id;

    /*!
     * \brief Holds the task's name.
     */
    std::string _name;

    /*!
     * \brief Holds the task's chain name.
     */
    std::string _chainName;

    /*!
     * \brief Holds the task's cost in seconds.
     */
    _float _cost;

    /*!
     * \brief Holds the task's deadline in seconds.
     */
    _float _deadline;

    /*!
     * \brief Holds the task's period in seconds.
     */
    _float _period;

    /*!
     * \brief Holds the task's priority level.
     */
    unsigned _priority;

    /*!
     * \brief Holds the offset jitter for the release of this task.
     */
    _float _offsetJitter;

    /*!
     * \brief Holds a reference to this task's traffic flow.
     */
    std::shared_ptr<Flow> _flow;

    /*!
     * \brief Holds the value for the response time for this application.
     * Note: If the response time has not been calculated before it holds the value -1.0.
     */
    _float _responseTime;

    /*!
     * \brief Calculates the response time for this task as described in:
     * \sa Audsley, N., Burns, A., Richardson, M., Tindell, K., & Wellings, A. J. (1993). Applying new scheduling theory to static priority pre-emptive scheduling. Software Engineering Journal, 8(5), 284. https://doi.org/10.1049/sej.1993.0034
     * \note It also saves the result to be used directly without a new calculation \sa getResponseTime.
     * \param periodFactor The period factor representing how fast is the processor executing this
     * task a parameter that influence the response time for this task.
     * \param higherPriorityTasks Vector of references to tasks that are executed in the same processor
     * as this task and have higher priorities.
     * \return The calculated value for the response time.
     */
    _float _calculateResponseTime(_float periodFactor,
                                  std::vector<std::shared_ptr<Task>> &higherPriorityTasks);

public:
    /*!
     * \brief Simple Constructor.
     */
    Task();

    /*!
     * \brief Complete Constructor.
     * \param name Task name.
     * \param chainName Task chain's name.
     * \param cost Task's cost.
     * \param deadline Tasks's deadline.
     * \param period Task's period.
     * \param priority Task's priority.
     * \param flow Task's traffic flow.
     */
    Task(std::string name, std::string chainName,
         _float cost, _float deadline, _float period, unsigned priority,
         const std::shared_ptr<Flow> &flow = nullptr);

    /*!
     * \brief Copy constructor.
     * \param other Other instance of Task.
     */
    Task(const Task &other);

    /*!
     * \brief Simple Destructor.
     */
    ~Task();

    /*!
     * \brief Getter to the task unique id number.
     * \return The unique identification number.
     */
    unsigned int getUniqueId();

    /*!
     * \brief Setter to the task priority.
     * \param priority The new priority level.
     */
    void setPriority(unsigned priority);

    /*!
     * \brief Getter to the task priority.
     * \return The task priority.
     */
    unsigned getPriority();

    /*!
     * \brief Setter to the task's name.
     * \param name The task new name.
     */
    void setName(const std::string name);

    /*!
     * \brief Getter to the task's name.
     * \return The task name.
     */

    const std::string getName();

    /*!
     * \brief Setter to task chain's name.
     * \param chainName The task chain's name.
     */
    void setChainName(const std::string chainName);

    /*!
     * \brief Getter to task chain's name.
     * \return The task chain's name.
     */
    const std::string getChainName();

    /*!
     * \brief Setter to the task cost.
     * \param cost The new cost value.
     */
    void setCost(_float cost);

    /*!
     * \brief Getter to the task cost.
     * \return The task cost.
     */
    _float getCost();

    /*!
     * \brief Setter to the task deadline.
     * \param deadline The new task deadline.
     */
    void setDeadline(_float deadline);

    /*!
     * \brief Getter to the task deadline.
     * \return The task deadline.
     */
    _float getDeadline();

    /*!
     * \brief Setter to the task period;
     * \param period The task period.
     */
    void setPeriod(_float period);

    /*!
     * \brief Getter to the task period.
     * \return The task period.
     */
    _float getPeriod();

    /*!
     * \brief Setter to the task offset release jitter.
     * \param offsetJitter The offset jitter value.
     */
    void setOffsetJitter(_float offsetJitter);

    /*!
     * \brief Getter to the task offset release jitter.
     * \return The offset release jitter value.
     */
    _float getOffsetJitter();

    /*!
     * \brief Setter to the task traffic flow.
     * \param flow The new traffic flow that has this task as source.
     */
    void setFlow(const std::shared_ptr<Flow> &flow);

    /*!
     * \brief Getter to this task traffic flow.
     * \return The traffic flow that has this task as source.
     */
    std::shared_ptr<Flow> getFlow() const;

    /*!
     * \brief Getter to this task response value.
     * \return Returns the value for this task response value. if this value has not been calculated
     * before it returns the value -1.0.
     */
    _float getResponseTime();

    /*!
     * \brief Calculates the response time for this task as described in:
     * \sa Audsley, N., Burns, A., Richardson, M., Tindell, K., & Wellings, A. J. (1993). Applying new scheduling theory to static priority pre-emptive scheduling. Software Engineering Journal, 8(5), 284. https://doi.org/10.1049/sej.1993.0034
     * \note It also saves the result to be used directly without a new calculation \sa getResponseTime.
     * \param sameProcTasks Vector of references to tasks that are executed in the same processor
     * as this task.
     * \return The calculated value for the response time.
     */
    _float calculateResponseTime(const std::vector<std::shared_ptr<Task>> &sameProcTasks);

    /*!
     * \brief Calculates the response time for this task as described in:
     * \sa Audsley, N., Burns, A., Richardson, M., Tindell, K., & Wellings, A. J. (1993). Applying new scheduling theory to static priority pre-emptive scheduling. Software Engineering Journal, 8(5), 284. https://doi.org/10.1049/sej.1993.0034
     * \note It also saves the result to be used directly without a new calculation \sa getResponseTime.
     * \param frequencyFactor The frequency factor for the processor executing this task. It indicates
     * how faster is the processor executing this task.
     * \param sameProcTasks Vector of references to tasks that are executed in the same processor
     * as this task.
     * \return The calculated value for the response time.
     */
    _float calculateScaledResponseTime(_float frequencyFactor, const std::vector<std::shared_ptr<Task>> &sameProcTasks);
};

#endif // TASK_H
